#include <stdio.h>
#include "mpi.h"
#include "mmpi.h"
#include "PLA.h"
#include "PLA_API.h"

/*
#define DEBUG	1
*/

static int PLA_State = PLA_STATE_NORMAL;

enum {
  PLA_API_ADD_SEND = 1 ,
  PLA_API_ADD_RECV = 2 };

/* ------------------------------------------------------------------ */
/* ------------------------------------------------------------------ */

int PLA_API_axpy_vector(
  MPI_Datatype datatype ,
  int          size ,
  void       * alpha ,
  void       * x ,
  int          xinc ,
  void       * y ,
  int          yinc )
{
  if      ( datatype == MPI_FLOAT ) {
    PLA_saxpy( &size, alpha, x, &xinc, y, &yinc );
  }
  else if ( datatype == MPI_DOUBLE ) {
    PLA_daxpy( &size, alpha, x, &xinc, y, &yinc );
  }
  return 0 ;
}

int PLA_API_axpy_matrix(
  MPI_Datatype datatype ,
  int          size_height ,
  int          size_width ,
  void       * alpha ,
  void       * x ,
  int          x_lda ,
  void       * y ,
  int          y_lda )
{
  int one = 1 ;

  if      ( datatype == MPI_FLOAT ) {
    float * xf = (float *) x ;
    float * yf = (float *) y ;
    int j ;
    for ( j = 0 ; j < size_width ; ++j ) {
      PLA_saxpy( &size_height , alpha , xf , &one , yf , &one ); /* JAG */
      xf += x_lda ;
      yf += y_lda ; /* JAG */
    }
  }
  else if ( datatype == MPI_DOUBLE ) {
    double * xf = (double *) x ;
    double * yf = (double *) y ;
    int j ;
    for ( j = 0 ; j < size_width ; ++j ) {
      PLA_daxpy( &size_height , alpha , xf , &one , yf , &one ); /* JAG */
      xf += x_lda ;
      yf += y_lda ; /* JAG */
    }
  }
  return 0 ;
}

/* ------------------------------------------------------------------ */
/* ------------------------------------------------------------------ */
/* !! MPI REQUIREMENT: Packing must match Unpacking exactly !! */

/* Pack from possibly strided memory */

static int PLA_API_vector_pack(
  int size , void * loc , int inc , MPI_Datatype datatype, int typesize,
  MMPI_Buf * Buf )
{
  MMPI_Buf_pack( &inc , 1 , MPI_INT , Buf );
  if ( inc == 1 ) {
    MMPI_Buf_pack( loc , size , datatype , Buf );
  }
  else {
    int k ;
    for ( k = 0 ; k < size ; ++k ) {
      MMPI_Buf_pack(((char *)loc) + k * inc * typesize, 1, datatype, Buf );
    }
  }
  return MPI_SUCCESS ;
}

/* Unpack into contiguous memory */

static int PLA_API_vector_unpack(
  MMPI_Buf Buf ,
  int size , void * loc,
  MPI_Datatype datatype , int typesize )
{
  int inc ;

  MMPI_Buf_unpack( Buf, &inc, 1, MPI_INT );
  if ( inc == 1 ) {
    MMPI_Buf_unpack( Buf , loc , size , datatype );
  }
  else { /* !!Unpacking to contiguous memory!! */
    int k ;
    for ( k = 0 ; k < size ; ++k ) {
      MMPI_Buf_unpack(Buf, ((char *)loc) + k * typesize, 1, datatype);
    }
  }
  return MPI_SUCCESS ;
}


static PLA_API_matrix_pack(
  int size_top, int size_left,
  void * loc , int lda ,
  MPI_Datatype datatype, int typesize,
  MMPI_Buf * Buf )
{
  int k ;
  for ( k = 0 ; k < size_left ; ++k ) {
    MMPI_Buf_pack(
      ((char *)loc) + k * lda * typesize , size_top, datatype, Buf );
  }
}

/* Unpack into contiguous memory !! => LDA == size_top */
static PLA_API_matrix_unpack(
  MMPI_Buf Buf ,
  int size_top , int size_left , void * loc ,
  MPI_Datatype datatype , int typesize )
{
  int k ;
  for ( k = 0 ; k < size_left ; ++k ) {
    MMPI_Buf_unpack( Buf , 
      ((char *)loc) + k * size_top * typesize , size_top , datatype );
  }
}


/* ------------------------------------------------------------------ */
/* ------------------------------------------------------------------ */

typedef struct {
  void       * buf ;     /* Receive buffer */
  void       * alpha ;   /* Scalar */
  int          num_row ; /* Number of rows */
  int          num_col ; /* Number of columns */
  void       * loc_p ;   /* pointer to local object */
  int          loc_ld ;  /* Leading dimension of local object */
  int          tag ;     /* Reserved tag */
  MPI_Comm     comm ;
  MPI_Datatype datatype ;
} PLA_API_Recv ;

int PLA_API_handler(
  void        * extra_state ,
  MPI_Request * req ,
  MPI_Status  * stat )
{
  PLA_API_Recv * data = (PLA_API_Recv *) extra_state ;

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"handler got: extra_state = 0x%x, buffer = 0x%x\n",
    data, data->buf);
#endif

  /* Release the local tag */

  MMPI_Tag_rel_local( data->comm , &(data->tag) );

  /* 'axpy' the result */

  if ( 0 == data->num_col ) {
    PLA_API_axpy_vector(
      data->datatype ,
      data->num_row ,
      data->alpha ,
      data->buf , 1 ,
      data->loc_p , data->loc_ld );
  }
  else {
    PLA_API_axpy_matrix(
      data->datatype ,
      data->num_row , data->num_col ,
      data->alpha ,
      data->buf ,   data->num_row ,
      data->loc_p , data->loc_ld );
  }  

  /* Free up all the storage */

  pla_free( data->buf );
  return(MPI_SUCCESS);
}

/* ------------------------------------------------------------------ */

int PLA_API_consumer( PLA_Obj laobj, int source, MMPI_Buf buffer )
/* 
 *  Purpose: handle api messages
 */
{
  const char name[] = "PLA_API_consumer" ;

  MPI_Comm         comm ;
  MPI_Datatype     datatype , mdatatype ;
  void           * cur_p , * tmp , * tmpa ;
  PLA_Obj_obj_type objtype;
  PLA_Obj          cur = NULL;
  int              curpos, curpos_row, curpos_col , cur_ld ;
  int              size, size_top, size_left ;
  int              typesize ;
  int              api_mode , api_tag , rem , k ;
                      
#ifdef DEBUG
  MMPI_Log_message(name,"begin");
#endif

  PLA_Obj_objtype( laobj, &objtype ); /* get object type */

  /* get data type information */

  PLA_Obj_datatype( laobj, &datatype );
  MPI_Type_size( datatype, &typesize );

  MMPI_Buf_comm( buffer , &comm );

  MMPI_Buf_unpack( buffer , &api_mode , 1 , MPI_INT );

  for (;;) {

    MMPI_Buf_remain( buffer , &rem );
    if ( 0 == rem ) break ;
    MMPI_Buf_unpack( buffer , &api_tag  , 1 , MPI_INT );


    switch( objtype ) {
    case PLA_VECTOR:

#ifdef DEBUG
  MMPI_Log_message(name,"vector");
#endif

      /* get location and size of where to put incoming data */
      MMPI_Buf_unpack( buffer, &curpos, 1, MPI_INT );
      MMPI_Buf_unpack( buffer, &size,   1, MPI_INT );

      /* take view of appropriate part of global object */
      PLA_Vector_subvector( laobj, size, curpos, &cur );
      PLA_Obj_local_buffer( cur , &cur_p );

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s subvector %d+%d @ 0x%x\n",name,curpos,size,cur_p);
  fflush(MMPI_Log_file());
#endif

      switch( api_mode ) {
      case PLA_API_ADD_SEND: /* Unpack and add */

        tmpa = (void *)pla_malloc( ( size + 1 ) * typesize );
        tmp  = ((char *)tmpa) + typesize ;

        MMPI_Buf_unpack( buffer, tmpa, 1, datatype );

        PLA_API_vector_unpack( buffer, size, tmp, datatype, typesize );

        PLA_API_axpy_vector( datatype , size , tmpa , tmp , 1, cur_p, 1 );

        pla_free( tmpa );

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s axpy\n",name);
  fflush(MMPI_Log_file());
#endif
        break ;

      case PLA_API_ADD_RECV: /* Pack & send */

        MPI_Rsend( cur_p, size, datatype, source, api_tag, comm );

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s reply -> %d @ %d\n",name,source,api_tag);
  fflush(MMPI_Log_file());
#endif
        break ;
      }
      PLA_Obj_free( &cur );
      break ;

    case PLA_MATRIX:

#ifdef DEBUG
  MMPI_Log_message(name,"matrix");
#endif

      /* get location of where to put incoming data */
      MMPI_Buf_unpack( buffer, &curpos_row, 1, MPI_INT );
      MMPI_Buf_unpack( buffer, &curpos_col, 1, MPI_INT );

      /* get size of incoming data */
      MMPI_Buf_unpack( buffer, &size_top, 1, MPI_INT );
      MMPI_Buf_unpack( buffer, &size_left, 1, MPI_INT );

      /* take view of appropriate part of global object */

      PLA_Matrix_submatrix( laobj, size_top, size_left, 
			    curpos_row, curpos_col, &cur );

      PLA_Obj_local_buffer( cur , &cur_p );
      PLA_Obj_local_ldim(   cur , &cur_ld );

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s submatrix %d+%d x %d+%d @ 0x%x : %d\n",
    name,curpos_row,size_top,curpos_col,size_left,cur_p,cur_ld);
  fflush(MMPI_Log_file());
#endif

      switch( api_mode ) {
      case PLA_API_ADD_SEND: /* Unpack and add */

        tmpa = (void *)pla_malloc( ( 1 + size_top * size_left ) * typesize );
        tmp  = (void *)(((char*)tmpa)+typesize);
        MMPI_Buf_unpack( buffer, tmpa, 1, datatype );
        PLA_API_matrix_unpack( buffer , size_top , size_left , tmp ,
          datatype , typesize );

        PLA_API_axpy_matrix(datatype,size_top,size_left,
          tmpa,tmp,size_top, cur_p,cur_ld);

        pla_free( tmpa );

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s axpy\n",name);
  fflush(MMPI_Log_file());
#endif

        break ;

      case PLA_API_ADD_RECV:

        /* PLA_Vector is always stride 1 */

        MPI_Type_vector( size_left, size_top, cur_ld, datatype, &mdatatype );
        MPI_Type_commit( &mdatatype );
        MPI_Rsend( cur_p, 1, mdatatype, source, api_tag, comm );
        MPI_Type_free( &mdatatype );

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s reply -> %d @ %d\n",name,source,api_tag);
  fflush(MMPI_Log_file());
#endif
        break ;
      }
      /* free view */
      PLA_Obj_free( &cur );
      break ;

    case PLA_MVECTOR:

#ifdef DEBUG
  MMPI_Log_message(name,"mvector");
#endif

      /* get location of where to put incoming data */
      MMPI_Buf_unpack( buffer, &curpos_row, 1, MPI_INT );
      MMPI_Buf_unpack( buffer, &curpos_col, 1, MPI_INT );

      /* get size of incoming data */
      MMPI_Buf_unpack( buffer, &size_top, 1, MPI_INT );
      MMPI_Buf_unpack( buffer, &size_left, 1, MPI_INT );

      /* take view of appropriate part of global object */
      PLA_Mvector_submvector( laobj, size_top, size_left, 
			      curpos_row, curpos_col, &cur );

      PLA_Obj_local_buffer(    cur , &cur_p );
      PLA_Obj_local_ldim( cur , &cur_ld );

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s submvector %d+%d x %d+%d @ 0x%x : %d\n",
    name,curpos_row,size_top,curpos_col,size_left,cur_p,cur_ld);
  fflush(MMPI_Log_file());
#endif

      switch( api_mode ) {
      case PLA_API_ADD_SEND:
        /* Unpack data */
        tmpa = (void *)pla_malloc( (1 + size_top * size_left) * typesize );
        tmp  = ( ((char *)tmpa) + typesize );
        MMPI_Buf_unpack( buffer, tmpa, 1, datatype );
        PLA_API_matrix_unpack( buffer , size_top , size_left , tmp ,
          datatype , typesize );
        PLA_API_axpy_matrix(datatype,size_top,size_left,
          tmpa,tmp,size_top, cur_p,cur_ld);
        pla_free( tmp );

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s axpy\n",name);
  fflush(MMPI_Log_file());
#endif

        break ;

      case PLA_API_ADD_RECV:

        MPI_Type_vector( size_left, size_top, cur_ld, datatype, &mdatatype );
        MPI_Type_commit( &mdatatype );
        MPI_Rsend( cur_p, 1, mdatatype, source, api_tag, comm );
        MPI_Type_free( &mdatatype );

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s reply -> %d @ %d\n",name,source,api_tag);
  fflush(MMPI_Log_file());
#endif

        break ;
      }
      /* free view */
      PLA_Obj_free( &cur );
      break ;
  
    default:
      return MPI_ERR_UNKNOWN ;
    }
  }

#ifdef DEBUG
  MMPI_Log_message(name,"END");
#endif

  return MPI_SUCCESS;
}

/* ------------------------------------------------------------------ */
/* ------------------------------------------------------------------ */

int PLA_Obj_API_mode( PLA_Obj laobj , int * state )
{
  *state = ( laobj->api != NULL ) ? PLA_STATE_API_ACTIVE : PLA_STATE_NORMAL ;
  return MPI_SUCCESS ;
}

int PLA_Obj_API_sync( PLA_Obj laobj )
{
  PLA_Template template ;
  int          num_proc , i ;
  MPI_Comm     comm ;
  MMPI_Con     consumer ;

  if ( ( consumer = (MMPI_Con) laobj->api ) == NULL ) {
    printf("Attempt to sync an object that is not API-ACTIVE\n");
    MPI_Abort( MPI_COMM_WORLD , MPI_ERR_ARG );
  } 

  PLA_Obj_template( laobj , &template );
  PLA_Temp_comm1D( template , &comm );
  PLA_Temp_total_num_proc( template , &num_proc );

  for ( i = 0 ; i < num_proc ; ++i ) MMPI_Con_wait( consumer , i );

  MMPI_Barrier( comm );

  return MPI_SUCCESS ;
}

int PLA_Obj_API_open( PLA_Obj laobj )
{
  const char name[] = "PLA_API_begin" ;
  MPI_Comm     comm ;
  MPI_Datatype datatype ;
  PLA_Template template ;

  if ( laobj->api ) {
    MMPI_Log_message( name , "API already active" );
    return MPI_ERR_UNKNOWN ;
  }

  PLA_Obj_template( laobj , &template );
  PLA_Temp_comm1D( template, &comm );

  /* DEBUGGING CODE */
  {
    int myproc, nproc , R ;
    if ( MPI_SUCCESS != ( R = MPI_Comm_size( comm , &myproc ) ) ) {
      MMPI_Log_abort( name , "PLA_Temp_comm1D gave bad communicator" , R );
    }
    if ( MPI_SUCCESS != ( R = MPI_Comm_rank( comm , &nproc ) ) ) {
      MMPI_Log_abort( name , "PLA_Temp_comm1D gave bad communicator" , R );
    }
  }

  return MMPI_Con_create( comm, laobj,
                          (MMPI_Con_handler) PLA_API_consumer,
                          (MMPI_Con *) &(laobj->api) );
}

int PLA_Obj_API_close( PLA_Obj laobj )
{
  const char name[] = "PLA_API_end" ;

  if ( laobj->api ) {
    MMPI_Con_free( (MMPI_Con *) &(laobj->api) );
    return MPI_SUCCESS ;
  }
  else {
    MMPI_Log_message( name , "API not active" );
    return MPI_ERR_UNKNOWN ;
  }
}

/* ------------------------------------------------------------------ */
/* ------------------------------------------------------------------ */

int PLA_API_multi_axpy_vector_to_global( 
    int     nsub ,
    int   * len ,
    void  * alpha ,
    void  * local_vector ,
    int     local_inc ,
    PLA_Obj laobj ,
    int   * displ )
/* 
 *  Purpose: add contents of local vector in global vector
 *           MMPI producer routine
 */
{
  const char name[] = "PLA_API_multi_axpy_vector_to_global" ;

  MMPI_Con         consumer = MMPI_CON_NULL ;
  MMPI_Buf       * buffer ;
  MPI_Datatype     datatype;
  PLA_Template     template;
  PLA_Obj_obj_type objtype ;

  PLA_Obj cur    = NULL ;
  PLA_Obj cur1   = NULL ;
  PLA_Obj curtop = NULL ;

  int api_mode = PLA_API_ADD_SEND ;
  int api_tag  = -1 ;

  int  gcurpos , lcurpos ;
  int  owner, nproc, me, typesize, size, i;
  void *curbuf;

  /* check if object is api-active */

  if ( ( consumer = (MMPI_Con) laobj->api ) == NULL ) {
    printf("Attempt to add-to object that is not API-ACTIVE\n");
    MPI_Abort( MPI_COMM_WORLD , MPI_ERR_ARG );
  } 

  PLA_Obj_objtype( laobj, &objtype );

  if ( objtype != PLA_VECTOR ) {
    printf("Attempt to fill an object that is not a vector with a vector\n");
    MPI_Abort( MPI_COMM_WORLD , MPI_ERR_ARG );
  }

  PLA_Obj_datatype( laobj, &datatype );
  PLA_Obj_template( laobj, &template );

  PLA_Temp_me1D( template, &me );
  PLA_Temp_total_num_proc( template , &nproc );

  MPI_Type_size( datatype, &typesize );

  buffer = (MMPI_Buf *) pla_malloc( sizeof(MMPI_Buf) * nproc );

  for ( i = 0 ; i < nproc ; ++i ) buffer[i] = MMPI_BUF_NULL ;

  lcurpos = 0 ;

  for ( i = 0 ; i < nsub ; ++i ) {

    gcurpos = displ[i] ;

    /* let cur = laobj( displ[i]:displ[i]+len[i]-1 ) */

    PLA_Vector_subvector( laobj, len[i], displ[i], &cur );

    for (;;) {
      /* split off subvector owned by one node */
      PLA_Obj_top( cur, &size, &owner );
      if ( size == 0 ) break;
      PLA_Obj_horz_split_2( cur, size, &curtop, &cur1 );

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s size %d, local %d, global %d\n",
    name,size,lcurpos,gcurpos);
  fflush(MMPI_Log_file());
#endif

      if ( owner == me ) {
        PLA_Obj_local_buffer( curtop , &curbuf );
        PLA_API_axpy_vector( datatype, size, alpha,
           ((char *)local_vector) + lcurpos * local_inc * typesize ,
           local_inc, curbuf,1);

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s local axpy\n",name);
  fflush(MMPI_Log_file());
#endif

      }
      else {
        if ( buffer[ owner ] == MMPI_BUF_NULL ) {
          /* initiate message to owner of curtop */
          MMPI_Con_init( consumer, buffer + owner );
          MMPI_Buf_pack( &api_mode, 1, MPI_INT, buffer + owner );
        }
        /* Pack the send request */
        MMPI_Buf_pack( &api_tag, 1, MPI_INT,  buffer + owner );
        MMPI_Buf_pack( &gcurpos, 1, MPI_INT,  buffer + owner );
        MMPI_Buf_pack( &size,    1, MPI_INT,  buffer + owner );
        MMPI_Buf_pack( alpha,    1, datatype, buffer + owner );
        PLA_API_vector_pack( size ,
          ((char *)local_vector) + lcurpos * local_inc * typesize ,
          local_inc , datatype , typesize , buffer + owner );
        }

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"%s remote axpy -> %d\n",name,owner);
  fflush(MMPI_Log_file());
#endif

/* ENHANCEMENT:
 * If buffer size > buffer send limit
 *   send buffer
 *   delete buffer
 *   set buffer to NULL
 * end if
 */

      gcurpos += size;
      lcurpos += size ;
      PLA_Obj_swap( &cur, &cur1 );
    }
  }

  PLA_Obj_free( &cur ); 
  PLA_Obj_free( &cur1 ); 
  PLA_Obj_free( &curtop ); 

  /* Send out the buffers */

  for ( i = 0 ; i < nproc ; ++i ) {
    if ( buffer[i] != MMPI_BUF_NULL ) {
      MMPI_Con_send( buffer[i] , i , consumer );
      MMPI_Buf_free( buffer + i );
    }
  }

  pla_free( buffer );

  MMPI_Serve_handler(); /* Opportunity for local consumer handler */
}

/* ------------------------------------------------------------------ */

int PLA_API_multi_axpy_matrix_to_global( 
    int     nsub_row ,
    int     nsub_col ,
    int   * num_row ,
    int   * num_col ,
    void  * alpha ,
    void  * local_matrix,
    int     local_matrix_ld ,
    PLA_Obj laobj ,
    int   * displ_row ,
    int   * displ_col )
/* 
 *  Purpose: add contents of local matrix to global matrix
 *           MMPI producer routine
 */
{
  const char name[] = "PLA_API_multi_axpy_matrix_to_global" ;
  MPI_Datatype datatype , mdatatype ;
  PLA_Template template;
  PLA_Obj    cur      = NULL,  cur1 = NULL,
             curleft  = NULL,  curleft1 = NULL,
             curtop   = NULL;

  MMPI_Buf       * buffer ;
  MMPI_Con         consumer ;
  MMPI_Con_handler handler ;

  int api_mode = PLA_API_ADD_SEND ;
  int api_tag  = -1 ;

  int objtype, typesize,
      nproc , dest, nprows, cur_ldim,
      gcurpos_row, gcurpos_col,
      lcurpos_row, lcurpos_col,
      lcur_row, lcur_col,
      size_top, size_left,
      owner_row, owner_col,
      myrow, mycol, me1d, i , j , k ;
  void *cur_p;

  /* check if object is active */
  if ( (consumer = (MMPI_Con) laobj->api ) == NULL ){
    printf("Attempt to fill object that is not API-ACTIVE\n");
    MPI_Abort( MPI_COMM_WORLD , MPI_ERR_ARG );
  }

  PLA_Obj_objtype(  laobj, &objtype );
  PLA_Obj_datatype( laobj, &datatype );
  PLA_Obj_template( laobj, &template );
  PLA_Temp_comm_col_rank(  template, &myrow );
  PLA_Temp_comm_row_rank(  template, &mycol );
  PLA_Temp_me1D(   template, &me1d );
  PLA_Temp_comm_col_size( template, &nprows );
  PLA_Temp_total_num_proc( template , &nproc );
  
  MPI_Type_size( datatype, &typesize );

  buffer = (MMPI_Buf *) pla_malloc( sizeof(MMPI_Buf) * nproc );

  for ( i = 0 ; i < nproc ; ++i ) buffer[i] = MMPI_BUF_NULL ;

  /* let cur = laobj( displ_row[i]:displ_row[i]+num_row[i]-1,
                      displ_col[j]:displ_col[j]+num_col[j]-1 ) */

  lcurpos_row = 0 ;

  for ( i = 0 ; i < nsub_row ; ++i ) {

    lcurpos_col = 0 ;

    for ( j = 0 ; j < nsub_col ; ++j ) {

      switch( objtype ) {
      case PLA_MATRIX:
        PLA_Matrix_submatrix(
          laobj, num_row[i], num_col[j], displ_row[i], displ_col[j], &cur );
        break ;
      case PLA_MVECTOR:
        PLA_Mvector_submvector(
          laobj, num_row[i], num_col[j], displ_row[i], displ_col[j], &cur );
        break ;
      default:
      printf("Attempt to fill an object that is not a matrix with a matrix\n");
        MPI_Abort( MPI_COMM_WORLD , MPI_ERR_ARG );
      }

      gcurpos_row = displ_row[i] ;
      lcur_row    = lcurpos_row ;

      for (;;) {
        /* split off top block owned by one node */
        PLA_Obj_top( cur, &size_top, &owner_row );
        if ( size_top == 0 ) break;
        PLA_Obj_horz_split_2( cur, size_top, &curtop, &cur1 );

        gcurpos_col = displ_col[j];
        lcur_col    = lcurpos_col ;

        if ( objtype == PLA_MATRIX ) {

          for(;;) {
            /* split off left panel owned by one column of nodes */
            PLA_Obj_left( curtop, &size_left, &owner_col );
            if ( size_left == 0 ) break;
            PLA_Obj_vert_split_2( curtop, size_left, &curleft, &curleft1 );

#ifdef DEBUG
  fprintf(MMPI_Log_file(),
    "%s submatrix size %d x %d, local %d x %d, global %d x %d\n", name ,
    size_top, size_left, lcur_row, lcur_col, gcurpos_row, gcurpos_col );
#endif

            if ( owner_row == myrow && owner_col == mycol ) {

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"  ...Local\n");
  fflush(MMPI_Log_file());
#endif

              PLA_Obj_local_buffer( curleft , &cur_p );
              PLA_Obj_local_ldim(   curleft , &cur_ldim );
              PLA_API_axpy_matrix(
                datatype , size_top, size_left, alpha ,
                ((char*)local_matrix) +
                  ( lcur_row + lcur_col * local_matrix_ld ) * typesize ,
                local_matrix_ld , cur_p , cur_ldim );
            }
            else {
	      /* initiate message to owner of curtop */

              dest = owner_col*nprows + owner_row;

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"  ...Remote -> %d\n", dest);
  fflush(MMPI_Log_file());
#endif

              if ( buffer[dest] == MMPI_BUF_NULL ) {
                MMPI_Con_init( consumer, buffer + dest );
                MMPI_Buf_pack( &api_mode, 1, MPI_INT, buffer + dest );
              }
              MMPI_Buf_pack( &api_tag,     1, MPI_INT, buffer + dest );
              MMPI_Buf_pack( &gcurpos_row, 1, MPI_INT, buffer + dest );
              MMPI_Buf_pack( &gcurpos_col, 1, MPI_INT, buffer + dest );
              MMPI_Buf_pack( &size_top,    1, MPI_INT, buffer + dest );
              MMPI_Buf_pack( &size_left,   1, MPI_INT, buffer + dest );
              MMPI_Buf_pack( alpha ,       1, datatype,buffer + dest );

              PLA_API_matrix_pack( size_top , size_left ,
                ((char*)local_matrix) +
                  ( lcur_row + lcur_col * local_matrix_ld ) * typesize ,
                local_matrix_ld ,
                datatype , typesize , buffer + dest );

/* ENHANCEMENT:
 *  if buffer size > buffer limit
 *    send buffer
 *    delete buffer
 *    set buffer = NULL
 *  end if
 */

            }

            gcurpos_col += size_left;
            lcur_col    += size_left;
            PLA_Obj_swap( &curtop, &curleft1 );
          }
        }
        else { /* Multivector */

#ifdef DEBUG
  fprintf(MMPI_Log_file(),
    "%s submatrix size %d x %d, local %d x %d, global %d x %d\n", name ,
    size_top, num_col[j], lcur_row, lcur_col, gcurpos_row, gcurpos_col );
  fflush(MMPI_Log_file());
#endif

          if ( owner_row == me1d ) {
            PLA_Obj_local_buffer( curtop , &cur_p );
            PLA_Obj_local_ldim(   curtop , &cur_ldim );

            PLA_API_axpy_matrix(
              datatype , size_top, num_col[j], alpha ,
              ((char *)local_matrix) +
                ( lcur_row + lcur_col * local_matrix_ld ) * typesize ,
              local_matrix_ld , cur_p , cur_ldim );
          }
          else {
            dest = owner_row ;
            if ( buffer[dest] == MMPI_BUF_NULL ) {
              MMPI_Con_init( consumer, buffer + dest );
              MMPI_Buf_pack( &api_mode, 1, MPI_INT, buffer + dest );
            }
    /* initiate message to owner of curtop */
            MMPI_Buf_pack( &api_tag,     1, MPI_INT, buffer + dest );
            MMPI_Buf_pack( &gcurpos_row, 1, MPI_INT, buffer + dest );
            MMPI_Buf_pack( &gcurpos_col, 1, MPI_INT, buffer + dest );
            MMPI_Buf_pack( &size_top,    1, MPI_INT, buffer + dest );
            MMPI_Buf_pack( num_col+j,    1, MPI_INT, buffer + dest );
            MMPI_Buf_pack( alpha,        1, datatype,buffer + dest );

            PLA_API_matrix_pack( size_top , size_left ,
              ((char*)local_matrix) +
                ( lcur_row + lcur_col * local_matrix_ld ) * typesize ,
              local_matrix_ld ,
              datatype , typesize , buffer + dest );

/* ENHANCEMENT:
 *  if buffer size > buffer limit
 *    send buffer
 *    delete buffer
 *    set buffer = NULL
 *  end if
 */

          }
        }

        gcurpos_row += size_top;
        lcur_row    += size_top;
        PLA_Obj_swap( &cur, &cur1 );
      }
      lcurpos_col += num_col[j] ;
    }
    lcurpos_row += num_row[i] ;
  }

  PLA_Obj_free( &cur ); 
  PLA_Obj_free( &curleft ); 
  PLA_Obj_free( &cur1 ); 
  PLA_Obj_free( &curleft1 ); 
  PLA_Obj_free( &curtop ); 

  for ( i = 0 ; i < nproc ; ++i ) {
    if ( buffer[i] != MMPI_BUF_NULL ) {
      MMPI_Con_send( buffer[i], i, consumer ); 
      MMPI_Buf_free( buffer + i );
    }
  }

  pla_free( buffer );

  MMPI_Serve_handler(); /* Opportunity for local consumer handler */
}

/* ------------------------------------------------------------------ */
/* ------------------------------------------------------------------ */

int PLA_API_multi_axpy_global_to_vector(
  int     nsub ,
  int   * len ,
  void  * alpha ,
  PLA_Obj laobj,
  int   * displ,
  void  * local_vector,
  int     local_inc )
/* 
 *
 *  Purpose: get contents of global vector and put in local vector
 *           MMPI producer routine
 */
{
  MPI_Comm     comm;
  MPI_Datatype datatype;
  MMPI_Con     consumer ;

  PLA_Obj cur    = NULL ;
  PLA_Obj cur1   = NULL ;
  PLA_Obj curtop = NULL;
  PLA_Template template;

  void *curtop_p ;
  int   objtype, size, owner, me, nproc, typesize, i ;
  int   lcurpos , gcurpos ;
  int   api_mode = PLA_API_ADD_RECV ;
  int   api_tag  = -1 ;

  /* check if object is fill-locked */

  if ( ( consumer = (MMPI_Con) laobj->api ) == NULL ) {
    printf("Attempt to query object that is not API-ACTIVE\n");
    MPI_Abort( MPI_COMM_WORLD , MPI_ERR_ARG );
  }; 

  PLA_Obj_objtype( laobj, &objtype );

  if ( objtype != PLA_VECTOR ) {
    printf("Attempt to fill an object that is not a vector with a vector\n");
    /* MPI_ERR_ABORT( MPI_ERR_ARG );  JAG */
    /* Sorry, is it supposed to be MPI_Abort with a missing arg or */
    /* an MMPI function that I couldn't find or ????               */
  }

  PLA_Obj_datatype( laobj, &datatype );
  PLA_Obj_template( laobj, &template );
  PLA_Temp_comm1D( template, &comm );
  PLA_Temp_me1D(   template, &me );
  PLA_Temp_total_num_proc( template , &nproc );

  MPI_Type_size( datatype, &typesize );

  lcurpos = 0 ;

  for ( i = 0 ; i < nsub ; ++i ) {

    gcurpos = displ[i];

    /* let cur = laobj( displ[i]:displ[i]+len[i]-1 ) */

    PLA_Vector_subvector( laobj, len[i], displ[i], &cur );

    for (;;) {
        /* split off subvector owned by one node */
      PLA_Obj_top( cur, &size, &owner );
      if ( size == 0 ) break;

/*printf("size = %d, owner = %d\n", size, owner); */

      PLA_Obj_horz_split_2( cur, size, &curtop, &cur1 );

      if ( owner == me ) {

/*printf("size = %d, owner = me\n", size, owner); */

        PLA_Obj_local_buffer( curtop , &curtop_p );
        PLA_API_axpy_vector(datatype,size,alpha,curtop_p,1,
          ((char *)local_vector) + lcurpos * local_inc * typesize,
          local_inc);
      }
      else {

        MPI_Request    req ;
        MMPI_Buf       mmpi_buf = MMPI_BUF_NULL ;
        void         * buf =
          (void *)pla_malloc( typesize * ( 1 + size ) + sizeof(PLA_API_Recv) );

        PLA_API_Recv * recv = (PLA_API_Recv *)
          (((char *)buf) + typesize*(1+size));

        recv->datatype = datatype ;

        MMPI_Tag_get_local( recv->comm = comm , &(recv->tag) );

        recv->buf     = buf ;
        recv->alpha   = (((char *)recv->buf) + typesize * size );
        recv->num_row = size ;
        recv->num_col = 0 ;
        recv->loc_p   =
          ((char *)local_vector) + lcurpos * local_inc * typesize;
        recv->loc_ld  = local_inc ;

        memcpy( recv->alpha , alpha , typesize );

        MPI_Irecv(
          recv->buf,
          recv->num_row,
          recv->datatype, owner,
          recv->tag,
          recv->comm , &(req) );

        MMPI_Post_handler( req , (void *) recv , PLA_API_handler );

        /* prepare message to owner of curtop */

        MMPI_Con_init( consumer, &mmpi_buf );
        MMPI_Buf_pack( &api_mode , 1, MPI_INT, &mmpi_buf );

        MMPI_Buf_pack( &(recv->tag),    1, MPI_INT, &mmpi_buf );
        MMPI_Buf_pack( &gcurpos,        1, MPI_INT, &mmpi_buf );
        MMPI_Buf_pack( &(recv->num_row),1, MPI_INT, &mmpi_buf );

        MMPI_Con_send( mmpi_buf , owner , consumer );
        /* MMPI_Con_wait( dest ); /* Wait for reply! */
        MMPI_Buf_free( &mmpi_buf );
      }
      gcurpos += size;
      lcurpos += size;
      PLA_Obj_swap( &cur, &cur1 );
    }
  }

  PLA_Obj_free( &cur ); 
  PLA_Obj_free( &cur1 ); 
  PLA_Obj_free( &curtop ); 

  MMPI_Serve_handler();     /* Opportunity for local consumer handler */
}

/* ------------------------------------------------------------------ */

int PLA_API_multi_axpy_global_to_matrix( 
    int     nsub_row ,
    int     nsub_col,
    int   * num_row ,
    int   * num_col ,
    void  * alpha ,
    PLA_Obj laobj,
    int   * displ_row,
    int   * displ_col,
    void  * local_matrix,
    int     local_matrix_ld )
/* 
 *  Purpose: add contents of local matrix to global matrix
 *           MMPI producer routine
 */
{
  MPI_Datatype datatype ;
  MPI_Comm     comm;
  MMPI_Con     consumer ;

  PLA_Template template;
  PLA_Obj      cur      = NULL,  
               curleft  = NULL,  
               curtop   = NULL;

  int          objtype, typesize,
               me, dest, nprows, nproc,
               base_row , base_col ,
               cur_row, cur_col,
               gcurpos_row, gcurpos_col,
               lcurpos_row, lcurpos_col,
               size_top, size_left,
               owner_row, owner_col,
               myrow, mycol, local_ldim, i , j ;
  int api_mode = PLA_API_ADD_RECV ;

  void         *cur_p ;

  /* check if object is fill-locked */
  if ( ( consumer = (MMPI_Con) laobj->api ) == NULL ) {
    printf("Attempt to fill object that is not FILL LOCKED\n");
    MPI_Abort( MPI_COMM_WORLD , MPI_ERR_ARG );
  }

  PLA_Obj_objtype(   laobj, &objtype );
  PLA_Obj_datatype(  laobj, &datatype );
  PLA_Obj_template(  laobj, &template );
  PLA_Obj_local_ldim( laobj, &local_ldim );

  PLA_Temp_comm1D( template, &comm);
  PLA_Temp_me1D(   template, &me);
  PLA_Temp_comm_col_rank(  template, &myrow );
  PLA_Temp_comm_row_rank(  template, &mycol );
  PLA_Temp_comm_col_size( template, &nprows );
  PLA_Temp_total_num_proc( template, &nproc );

  MPI_Type_size( datatype, &typesize );

  base_row = 0 ;
  for ( i = 0 ; i < nsub_row ; i++ ) {
    base_col = 0 ;
    for ( j = 0 ; j < nsub_col ; j++ ) {

      /* let cur = laobj( displ_row[i]:displ_row[i]+num_row[i]-1,
                          displ_col[j]:displ_col[j]+num_col[j]-1 ) */

      lcurpos_row = base_row ;
      gcurpos_row = displ_row[i];

      switch( objtype ) {
      case PLA_MATRIX:
        PLA_Matrix_submatrix(
          laobj, num_row[i], num_col[j], displ_row[i], displ_col[j], &cur );
        break ;

      case PLA_MVECTOR:
        PLA_Mvector_submvector(
          laobj, num_row[i], num_col[j], displ_row[i], displ_col[j], &cur );
        break ;

      default:
      printf("Attempt to fill an object that is not a matrix with a matrix\n");
        MPI_Abort( MPI_COMM_WORLD , MPI_ERR_ARG );
      }

      for (;;) {
        /* split off top block owned by one node */
        PLA_Obj_top( cur, &size_top, &owner_row );
        if ( size_top == 0 ) break;
        PLA_Obj_horz_split_2( cur, size_top, &curtop, &cur );
        lcurpos_col = base_col ;
        gcurpos_col = displ_col[j] ;

        if ( objtype == PLA_MATRIX ) {

          for (;;) {
            /* split off left panel owned by one column of nodes */

            PLA_Obj_left( curtop, &size_left, &owner_col );
            if ( size_left == 0 ) break;
            PLA_Obj_vert_split_2( curtop, size_left, &curleft, &curtop );

            PLA_Obj_local_buffer( curleft , &cur_p );

            dest = owner_col*nprows + owner_row;

            if ( dest == me ) {
              /* local add */
              PLA_API_axpy_matrix(
                datatype , size_top , size_left , alpha ,
                cur_p , local_ldim ,
                ((char *)local_matrix) +
                  ( lcurpos_row + lcurpos_col * local_matrix_ld ) * typesize,
                local_matrix_ld );
            }
            else {

              MPI_Request req ;
              MMPI_Buf    mmpi_buf = MMPI_BUF_NULL ;
              void      * buf = (void *)pla_malloc(
                 typesize * (1 + size_top * size_left) + sizeof(PLA_API_Recv));

              PLA_API_Recv * recv = (PLA_API_Recv *)
                (((char *)buf) + typesize * ( 1 + size_top * size_left ));

              recv->datatype = datatype ;

              MMPI_Tag_get_local( recv->comm = comm , &(recv->tag) );

              recv->buf     = buf ;
              recv->alpha   = (((char *)recv->buf) +
                                typesize * size_top * size_left );
              recv->num_row = size_top ;
              recv->num_col = size_left ;
              recv->loc_p   =
                ((char *)local_matrix) +
                  ( lcurpos_row + lcurpos_col * local_matrix_ld ) * typesize ;

              recv->loc_ld  = local_matrix_ld ;

              memcpy( recv->alpha , alpha , typesize );

              MPI_Irecv(
                recv->buf,
                recv->num_row * recv->num_col ,
                recv->datatype, dest,
                recv->tag,
                recv->comm , &(req) );

              MMPI_Post_handler( req , (void *) recv , PLA_API_handler );

#ifdef DEBUG
  fprintf(MMPI_Log_file(),"reply to: extra_state = 0x%x, buffer = 0x%x\n",
    recv, recv->buf);
#endif

              /* prepare message to owner of curtop */

              MMPI_Con_init( consumer, &mmpi_buf );
              MMPI_Buf_pack( &api_mode , 1, MPI_INT, &mmpi_buf );

              MMPI_Buf_pack( &(recv->tag), 1, MPI_INT, &mmpi_buf );
              MMPI_Buf_pack( &gcurpos_row, 1, MPI_INT, &mmpi_buf );
              MMPI_Buf_pack( &gcurpos_col, 1, MPI_INT, &mmpi_buf );
              MMPI_Buf_pack( &size_top,    1, MPI_INT, &mmpi_buf );
              MMPI_Buf_pack( &size_left,   1, MPI_INT, &mmpi_buf );

              MMPI_Con_send( mmpi_buf , dest , consumer );
              /* MMPI_Con_wait( dest ); /* Wait for reply! */
              MMPI_Buf_free( &mmpi_buf );
            }

            lcurpos_col += size_left ;
            gcurpos_col += size_left ;
          }
        }
        else if ( objtype == PLA_MVECTOR ) {
          size_left = num_col[j] ;
          dest = owner_row ;
          PLA_Obj_local_buffer( curtop , &cur_p );

          if ( dest == me ) {
            /* local add */
            PLA_API_axpy_matrix(
              datatype , size_top , size_left , alpha ,
              cur_p , local_ldim ,
              ((char *)local_matrix) +
                ( lcurpos_row + lcurpos_col * local_matrix_ld ) * typesize ,
              local_matrix_ld );
          }
          else {

              MPI_Request req ;
              MMPI_Buf    mmpi_buf = MMPI_BUF_NULL ;
              void      * buf = (void *)pla_malloc(
                 typesize * (1 + size_top * size_left) + sizeof(PLA_API_Recv));

              PLA_API_Recv * recv = (PLA_API_Recv *)
                (((char *)buf) + typesize * ( 1 + size_top * size_left ));

              recv->datatype = datatype ;
              MMPI_Tag_get_local( recv->comm = comm , &(recv->tag) );

              recv->buf     = buf ;
              recv->alpha   = (((char *)recv->buf) +
                                typesize * size_top * size_left );
              recv->num_row = size_top ;
              recv->num_col = size_left ;
              recv->loc_p   =
                ((char *)local_matrix) +
                  ( lcurpos_row + lcurpos_col * local_matrix_ld ) * typesize ,
              recv->loc_ld  = local_matrix_ld ;

              memcpy( recv->alpha , alpha , typesize );

              MPI_Irecv(
                recv->buf,
                recv->num_row * recv->num_col ,
                recv->datatype, dest,
                recv->tag,
                recv->comm , &(req) );

              MMPI_Post_handler( req , (void *) recv , PLA_API_handler );

              /* prepare message to owner of curtop */

              MMPI_Con_init( consumer, &mmpi_buf );
              MMPI_Buf_pack( &api_mode , 1, MPI_INT, &mmpi_buf );

              MMPI_Buf_pack( &(recv->tag), 1, MPI_INT, &mmpi_buf );
              MMPI_Buf_pack( &gcurpos_row, 1, MPI_INT, &mmpi_buf );
              MMPI_Buf_pack( &gcurpos_col, 1, MPI_INT, &mmpi_buf );
              MMPI_Buf_pack( &size_top,    1, MPI_INT, &mmpi_buf );
              MMPI_Buf_pack( &size_left,   1, MPI_INT, &mmpi_buf );

              MMPI_Con_send( mmpi_buf , dest , consumer );
              /* MMPI_Con_wait( dest ); /* Wait for reply! */
              MMPI_Buf_free( &mmpi_buf );
          }
        }
        gcurpos_row += size_top;
        lcurpos_row += size_top ;
      }

      base_col += num_col[j] ;
    }
    base_row += num_row[i] ;
  }

  PLA_Obj_free( &cur ); 
  PLA_Obj_free( &curtop ); 
  PLA_Obj_free( &curleft ); 

  MMPI_Serve_handler();     /* Opportunity for local consumer handler */
}


/* ------------------------------------------------------------------ */
/* ------------------------------------------------------------------ */
/* Add local vector to global subvector */

int PLA_API_axpy_vector_to_global(
  int     size ,      /* Length */
  void  * alpha ,     /* Scalar */
  void  * local_obj , /* Local vector values */
  int     local_inc , /* Local vector increment */
  PLA_Obj laobj ,     /* Global vector */
  int     disp )      /* Global vector displacement */
{
  return PLA_API_multi_axpy_vector_to_global(
    1, &size, alpha, local_obj, local_inc, laobj, &disp );
}

/* Add global subvector to local vector */

int PLA_API_axpy_global_to_vector(
  int     size ,
  void  * alpha ,
  PLA_Obj laobj ,
  int     disp ,
  void  * local_obj ,
  int     local_inc )
{
  return PLA_API_multi_axpy_global_to_vector(
    1, &size, alpha, laobj, &disp, local_obj, local_inc );
}

/* Add local matrix to global (sub)matrix or (sub)multivector */

int PLA_API_axpy_matrix_to_global(
  int     size_row ,
  int     size_col ,
  void  * alpha ,
  void  * local_obj ,
  int     local_ldim ,
  PLA_Obj laobj ,
  int     disp_row ,
  int     disp_col )
{
  return PLA_API_multi_axpy_matrix_to_global(
    1, 1, &size_row, &size_col, alpha, local_obj, local_ldim ,
    laobj, &disp_row, &disp_col );
}

/* Add global (sub)matrix or (sub)multivector to local matrix */

int PLA_API_axpy_global_to_matrix(
  int     size_row ,
  int     size_col ,
  void  * alpha ,
  PLA_Obj laobj ,
  int     disp_row ,
  int     disp_col ,
  void  * local_obj ,
  int     local_ldim )
{
  return PLA_API_multi_axpy_global_to_matrix(
    1, 1, &size_row, &size_col, alpha,
    laobj, &disp_row, &disp_col, local_obj, local_ldim );


}
int PLA_API_begin()
{
   MPI_Comm  base_comm;

/*    PLA_Base_comm_1d(&base_comm);
   MMPI_Enable(base_comm); */
   PLA_State = PLA_STATE_API_ACTIVE;
}


int PLA_API_end()
{
   MPI_Comm  base_comm;

/*   PLA_Base_comm_1d(&base_comm);
   MMPI_Disable(base_comm); */
   PLA_State = PLA_STATE_NORMAL;
}

int PLA_API_state(int *state)
{
   *state = PLA_State;
   return PLA_SUCCESS;
}

