
/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/



#include "PLA.h"
#include <stdio.h>

int PLA_me, PLA_np, PLA_nprows, PLA_npcols;
int PLA_i_one=1, PLA_SYNC, PLA_flow_direction_x, PLA_flow_direction_y;
MPI_Comm PLA_row, PLA_col;
double PLA_d_one=1.0;
/********************************************************************

All of the routines in this file are used to implement a pipelined
ring broadcast strictly within a set of rows or columns (see the
MPI calls referring to creating a cartesian grid in PLA_init.c).

MPI is relied on to do the buffering and the only reduce operation
implemented is summing (because that is all our implementation needs).


PLA_Bcast()    : Determines the direction of the broadcast based on
                the globals set via PLA_Set_Flow and the MPI_Comm used.
                It then calls one of the four (one for each 2-D direction)
                pipeline broadcast routines.

PLA_Reduce2()   : Similar to Bcast except that a reduce is done in the
                indicated direction.  To be more accurate -- a SUM is
                done (via MPI_Sum) in the appropriate direction.

PLA_Set_Flow() : Simply sets the direction of the flow within the
                processor mesh (UP/DOWN, RIGHT/LEFT) and the
                sync property to be either ASYNCHRONOUS (use ring
                broadcast or SYNCHRONOUS (defualt to MPI's tree
                broadcast structure).

Programmed by John A. Gunnels
              gunnels@cs.utexas.edu

Copyright 1996  The University of Texas at Austin
This code is distributed under gnu license.
********************************************************************/
void UP_Bcast(double* buf,int count,MPI_Datatype datatype,
                 int iroot,MPI_Comm comm )
{
  MPI_Status status;
  int me;
  MPI_Comm_rank( comm, &me );
  if ( me != iroot )
    MPI_Recv( buf, count, datatype,(me+1)%PLA_nprows,MPI_ANY_TAG,comm,&status);
  if( (me-1+PLA_nprows)%PLA_nprows != iroot)
     MPI_Ssend(buf, count, datatype, (me-1+PLA_nprows)%PLA_nprows,0, comm);
}
/************************************************************************/
void UP_Sum(void *buf,int count,MPI_Datatype type,int root,MPI_Comm comm,
            void *work)
{
    MPI_Status status;
    int me;
    MPI_Comm_rank( comm, &me );
    if(me != (root-1+PLA_nprows)%PLA_nprows){
        MPI_Recv(work,count,type,(me+1)%PLA_nprows,MPI_ANY_TAG,comm,&status);
        PLA_daxpy(&count,&PLA_d_one,buf,&PLA_i_one,work,&PLA_i_one);
    }
    else PLA_dcopy(&count,buf,&PLA_i_one,work,&PLA_i_one);

    if(me !=root)
        MPI_Ssend(work,count,type, (me-1+PLA_nprows)%PLA_nprows ,0,comm);
}
/************************************************************************/
void DOWN_Bcast(void *buf,int count, MPI_Datatype datatype,int iroot,
                MPI_Comm comm )
{
  MPI_Status status;
  int me;

  MPI_Comm_rank( comm, &me );
  if ( me != iroot )
    MPI_Recv( buf, count, datatype, (me-1+PLA_nprows)%PLA_nprows,
              MPI_ANY_TAG, comm, &status );
  if ( (me+1)%PLA_nprows != iroot)
     MPI_Ssend(buf, count, datatype, (me+1)%PLA_nprows,0, comm);
}
/************************************************************************/
void DOWN_Sum(void *buf,int count,MPI_Datatype type,int root,MPI_Comm comm,
            void *work)
{
    MPI_Status status;
    int me;

    MPI_Comm_rank( comm, &me );
    if(me != (root+1)%PLA_nprows){
        MPI_Recv(work,count,type,(me-1+PLA_nprows)%PLA_nprows,MPI_ANY_TAG,comm,&status);
        PLA_daxpy(&count,&PLA_d_one,buf,&PLA_i_one,work,&PLA_i_one);
    }
    else PLA_dcopy(&count,buf,&PLA_i_one,work,&PLA_i_one);
    if(me !=root)
        MPI_Ssend(work,count,type,(me+1)%PLA_nprows,0,comm);
}
/************************************************************************/
void LEFT_Bcast(double* buf,int count,MPI_Datatype datatype,
                 int iroot,MPI_Comm comm )
{
  MPI_Status status;
  int me;

  MPI_Comm_rank( comm, &me );
  if ( me != iroot )
    MPI_Recv( buf, count, datatype,(me+1)%PLA_npcols,MPI_ANY_TAG,comm,&status);
  if( (me-1+PLA_npcols)%PLA_npcols != iroot)
     MPI_Ssend(buf, count, datatype, (me-1+PLA_npcols)%PLA_npcols,0, comm);
}
/************************************************************************/
void LEFT_Sum(void *buf,int count,MPI_Datatype type,int root,MPI_Comm comm,
            void *work)
{
    MPI_Status status;
    int me;

    MPI_Comm_rank( comm, &me );
    if(me != (root-1+PLA_npcols)%PLA_npcols){
        MPI_Recv(work,count,type,(me+1)%PLA_npcols,MPI_ANY_TAG,comm,&status);
        PLA_daxpy(&count,&PLA_d_one,buf,&PLA_i_one,work,&PLA_i_one);
    }
    else PLA_dcopy(&count,buf,&PLA_i_one,work,&PLA_i_one);
    if(me !=root)
        MPI_Ssend(work,count,type, (me-1+PLA_npcols)%PLA_npcols ,0,comm);
}
/************************************************************************/
void RIGHT_Bcast(double* buf,int count,MPI_Datatype datatype,
                 int iroot,MPI_Comm comm )
{
  MPI_Status status;
  int me;

  MPI_Comm_rank( comm, &me );
  if ( me != iroot )
    MPI_Recv( buf, count, datatype, (me-1+PLA_npcols)%PLA_npcols,
              MPI_ANY_TAG, comm, &status );
  if ( (me+1)%PLA_npcols != iroot)
     MPI_Ssend(buf, count, datatype, (me+1)%PLA_npcols,0, comm);
}
/************************************************************************/
void RIGHT_Sum(void *buf,int count,MPI_Datatype type,int root,MPI_Comm comm,
            void *work)
{
    MPI_Status status;
    int me;

    MPI_Comm_rank( comm, &me );
    if(me != (root+1)%PLA_npcols){
        MPI_Recv(work,count,type,(me-1+PLA_npcols)%PLA_npcols,MPI_ANY_TAG,comm,&status);
        PLA_daxpy(&count,&PLA_d_one,buf,&PLA_i_one,work,&PLA_i_one);
    }
    else PLA_dcopy(&count,buf,&PLA_i_one,work,&PLA_i_one);
    if(me !=root)
        MPI_Ssend(work,count,type,(me+1)%PLA_npcols,0,comm);
}
/************************************************************************/
int PLA_Reduce2(void *source,void *sink,int length,MPI_Datatype type, 
            MPI_Op operation,int root, MPI_Comm comm)
{
   int row, col;

   if(PLA_SYNC == SYNCHRONOUS)
   {
      MPI_Reduce(source, sink, length, type, operation, root, comm);
   }
   else
   {
      MPI_Comm_compare(PLA_row, comm, &row);
      MPI_Comm_compare(PLA_col, comm, &col);

      if((PLA_flow_direction_x == RIGHT) && (row == MPI_IDENT))
      {
         RIGHT_Sum(source, length, type, root, comm, sink);
      }
      else if((PLA_flow_direction_x == LEFT) && (row == MPI_IDENT))
      {
         LEFT_Sum(source, length, type, root, comm, sink);
      }
      else if((PLA_flow_direction_y == UP) && (col == MPI_IDENT))
      {
         UP_Sum(source, length, type, root, comm, sink);
      }
      else if((PLA_flow_direction_y == DOWN) && (col == MPI_IDENT))
      {
         DOWN_Sum(source, length, type, root, comm, sink);
      }
   }
   return 0;
}

int PLA_Bcast(void *source, int length, MPI_Datatype type,int root,
           MPI_Comm comm)
{
   int row, col;

   if(PLA_SYNC == SYNCHRONOUS)
   {
      MPI_Bcast(source, length, type, root, comm);
   }
   else
   {
      MPI_Comm_compare(PLA_row, comm, &row);
      MPI_Comm_compare(PLA_col, comm, &col);

      if((PLA_flow_direction_x == RIGHT) && (row == MPI_IDENT))
{
         RIGHT_Bcast(source, length, type, root, comm);
}
      else if((PLA_flow_direction_x == LEFT) && (row == MPI_IDENT))
         LEFT_Bcast(source, length, type, root, comm);
      else if((PLA_flow_direction_y == UP) && (col == MPI_IDENT))
         UP_Bcast(source, length, type, root, comm);
      else if((PLA_flow_direction_y == DOWN) && (col == MPI_IDENT))
         DOWN_Bcast(source, length, type, root, comm);
      else printf("All of the comparisons failed -- try again bozo\n");
   }
   return 0;
}

int PLA_Set_Flow(int sync_type,int x_direction,int y_direction)
{
   if((sync_type == SYNCHRONOUS) || (sync_type == ASYNCHRONOUS))
      PLA_SYNC = sync_type;
   if((x_direction == RIGHT) || (x_direction == LEFT))
      PLA_flow_direction_x = x_direction;
   if((y_direction == DOWN) || (y_direction == UP))
      PLA_flow_direction_y = y_direction;
   return 0;
}

