/* ---------------------------------------------------------------------
Author:     H. Carter Edwards
		http://www.ticam.utexas.edu/~carter
		carter@ticam.utexas.edu

Copyright:  Copyright (C) 1996   H. Carter Edwards

Purpose:    Asynchronous message management for MPI

License:    See 'License.Gnu'

Mail:
	University of Texas at Austin
	TICAM, TAY 2.400
	2313 Speedway
	Austin, TX  78712

Disclaimer:

    MMPI comes with ABSOLUTELY NO WARRANTY; for details read
    the file License.Gnu that came with this library.
    This is free software, and you are welcome to redistribute it
    under certain conditions, i.e. read the file 'License.Gnu'.
--------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include "mmpi.h"

#define GLEN	100
#define LLEN	10

extern int con_handler( void * , int , MMPI_Buf );
extern int subvec_get(
  MMPI_Con consumer , int * disp , int * len , double * val );

main( int argc , char * argv[] )
{
  MMPI_Con consumer    = MMPI_CON_NULL ;
  double   local_data[  GLEN ];
  double   global_data[ GLEN ];
  double   recv_data[ LLEN ];
  int      i , proc , disp , len ;

  MPI_Init( &argc, &argv );   /* Initialize MPI */
  MMPI_Log_init( "Demo2.P" ); /* Name the MMPI Error/Warning file */
  MMPI_Enable( MPI_COMM_WORLD );/* Initialize MMPI on this communicator */

  MPI_Comm_rank( MPI_COMM_WORLD , &proc );

  /* Initialize vector */

  for ( i = 0 ; i < GLEN ; ++i ) local_data[i] = 0.0 ;
  for ( i = 0 ; i < GLEN ; ++i ) global_data[i] = 1000 * ( proc + 1 ) + i ;

  /* Synchronous creation of Conusumer object.
     The 'extra_state' is the shared vector 'global_data'.
  */

  MMPI_Con_create( MPI_COMM_WORLD, global_data, con_handler, &consumer );

  /* Begin asynchronous loop */

  while ( subvec_get( consumer, &disp, &len, recv_data ) ) {

    /* Sum received subvector */

    for ( i = 0 ; i < len ; ++i ) local_data[disp+i] += recv_data[i] ;
  }

  /* End asynchronous loop. */

  MMPI_Con_free( &consumer ); /* MMPI_Barrier called internally */

  /* Print the results */

  for ( i = 0 ; i < GLEN ; ++i ) {
    if ( 0 == i % 5 ) fprintf( MMPI_Log_file() , "\n" );
    fprintf( MMPI_Log_file() , " %f" , local_data[i] );
  }
  fprintf( MMPI_Log_file() , "\n" );

  MMPI_Disable( MPI_COMM_WORLD );
  MPI_Finalize();

  return 0 ;
}

/* Request random subvector contributions */

int subvec_get( MMPI_Con consumer , int * disp , int * len , double * val )
{
  static int count = 0 ;
  static int num   = 0 ;
  static int proc  = -1 ;
  static int nproc = -1 ;
  MPI_Request request ;
  MPI_Status  status ;
  MMPI_Buf send_buffer = MMPI_BUF_NULL ;
  int src , tag ;

  if ( 0 == num ) {
    MPI_Comm_size( MPI_COMM_WORLD , &nproc );
    MPI_Comm_rank( MPI_COMM_WORLD , &proc );

    srand( proc ); /* Different random numbers on each processor */

    num = 10 + rand() % 10 ; /* Random number of iterations */
  }

  *disp = rand() % ( GLEN - LLEN );  /* Random displacement */
  *len  = 1 + rand() % ( LLEN - 1 ); /* Random length */
   src  = rand() % nproc ;           /* Random source */

  MMPI_Tag_get_local( MPI_COMM_WORLD , &tag );

  MPI_Irecv( val, *len, MPI_DOUBLE, src, tag, MPI_COMM_WORLD, &request );

  MMPI_Con_init( consumer , &send_buffer );
  MMPI_Buf_pack( &tag , 1 , MPI_INT , &send_buffer );
  MMPI_Buf_pack( disp , 1 , MPI_INT , &send_buffer );
  MMPI_Buf_pack( len  , 1 , MPI_INT , &send_buffer );
  MMPI_Con_send( send_buffer , src , consumer );

  MMPI_Wait( &request, &status );

  MMPI_Tag_rel_local( MPI_COMM_WORLD , &tag );

  return ( ++count < num );
}

/* Subvector query handler */

int con_handler( void * extra_state , int source , MMPI_Buf Buf )
{
  double * global_data = (double *) extra_state ;
  MPI_Comm comm ;
  int tag , disp , len ;

  /* Unpack the request */

  MMPI_Buf_unpack( Buf , &tag  , 1 , MPI_INT );
  MMPI_Buf_unpack( Buf , &disp , 1 , MPI_INT );
  MMPI_Buf_unpack( Buf , &len ,  1 , MPI_INT );

  /* Reply */

  MMPI_Buf_comm( Buf , &comm );

  MPI_Rsend( global_data + disp , len , MPI_DOUBLE , source , tag , comm );

  return MPI_SUCCESS ;
}

