/* ---------------------------------------------------------------------
Author:     H. Carter Edwards
		http://www.ticam.utexas.edu/~carter
		carter@ticam.utexas.edu

Copyright:  Copyright (C) 1996   H. Carter Edwards

Purpose:    Asynchronous message management for MPI

License:    See 'License.Gnu'

Mail:
	University of Texas at Austin
	TICAM, TAY 2.400
	2313 Speedway
	Austin, TX  78712

Disclaimer:

    MMPI comes with ABSOLUTELY NO WARRANTY; for details read
    the file License.Gnu that came with this library.
    This is free software, and you are welcome to redistribute it
    under certain conditions, i.e. read the file 'License.Gnu'.
--------------------------------------------------------------------- */

#include <stdlib.h>
#include "mmpi.h"

/*
#define DEBUG
*/

/* ------------------------------------------------------------------ */

#define MINIMUM_BUFFER_SIZE	128

#define HEADER_LEN	2
#define HEADER_WHICH	0
#define HEADER_SIZE	1

#define REQUEST_CREATE	0x01
#define REQUEST_RESIZE	0x02
#define REQUEST_DESTROY	0x03
#define REQUEST_MESSAGE	0x10

typedef struct MMPI_Con_type {
  int         * send_size ;
  MMPI_Buf    * recv_buffer ;
  MPI_Request * request ;
  MPI_Request * req_ack ;
  MPI_Request   req_mgr ;
  MPI_Comm      comm ;
  int           tag_mgr ;
  int           tag_con ;
  int           tag_ack ;
  void        * extra_state ;
  MMPI_Con_handler handler;
  int           header[ HEADER_LEN ];
} MMPI_Con_type ;

/* ------------------------------------------------------------------ */

int MMPI_Con_verify( MMPI_Con con )
{
  const char name[] = "MMPI_Con_verify" ;
  int R , i ;
  int nproc ;

  /* Verify creation */

  if ( MPI_SUCCESS != ( R = MPI_Comm_size( con->comm , &nproc ) ) ) {
    MMPI_Log_abort( name , "MMPI_Comm_size failed" , R );
    return MPI_ERR_COMM ;
  }

  if ( ( con->send_size   != (int *)         ( con + 1 ) ) ||
       ( con->recv_buffer != (MMPI_Buf *)    ( con->send_size   + nproc ) ) ||
       ( con->request     != (MPI_Request *) ( con->recv_buffer + nproc ) ) ||
       ( con->req_ack     != (MPI_Request *) ( con->request     + nproc ) ) ) {
    MMPI_Log_abort( name , "corrupted pointers?" , MPI_ERR_UNKNOWN );
    return MPI_ERR_UNKNOWN ;
  }

  /* Verify buffers */

  for ( i = 0 ; i < nproc ; ++i ) {
    if ( con->recv_buffer[i] != MMPI_BUF_NULL ) {
      if ( MPI_SUCCESS != ( R = MMPI_Buf_verify( con->recv_buffer[i] ) ) ) {
        MMPI_Log_abort( name, "corrupted recv buffer", R );
        return R ;
      }
    }
  }

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------ */

int MMPI_Con_comm( MMPI_Con Con , MPI_Comm * Comm )
{
  return Con ? ( ( *Comm = Con->comm ) , MPI_SUCCESS ) : MPI_ERR_ARG ;
}

int MMPI_Con_func( MMPI_Con Con , MMPI_Con_handler * func )
{
  return Con ? ( ( *func = Con->handler ) , MPI_SUCCESS ) : MPI_ERR_ARG ;
}

int MMPI_Con_data( MMPI_Con Con , void ** data )
{
  return Con ? ( ( *data = Con->extra_state ) , MPI_SUCCESS ) : MPI_ERR_ARG ;
}

/* ------------------------------------------------------------------ */
/* The Con handler passes incoming messaged to the application */

static int con_handler(
  void        * extra_state ,
  MPI_Request * request ,
  MPI_Status  * status )
{
  const char my_name[] = "(MMPI)con_handler" ;
  int      source = status->MPI_SOURCE ;
  MMPI_Con con    = (MMPI_Con) extra_state ;
  int R , H ;

  if ( con->request[ source ] != *request ) {
    MMPI_Log_abort(my_name,"Corrupted Request", MPI_ERR_REQUEST);
  }

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : from %d, buf 0x%x\n",my_name,source,
    con->recv_buffer[source] );
  fflush(MMPI_Log_file());

#endif

  if( MPI_SUCCESS != ( R = MMPI_Buf_status(con->recv_buffer[source],status))){
    MMPI_Log_abort(my_name,"Failed MPI_Buf_status", R);
  }

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : Unpacking\n",my_name);
  fflush(MMPI_Log_file());

#endif

  if ( MPI_SUCCESS != ( R = MMPI_Buf_unpack(
    con->recv_buffer[source],&H,1,MPI_INT) ) ) {
    MMPI_Log_abort(my_name,"Failed MMPI_Buf_unpack", R);
  }

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : header(%d)\n",my_name,H);
  fflush(MMPI_Log_file());

#endif

  switch( H ) {
  case REQUEST_RESIZE:

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : Resize\n",my_name);
  fflush(MMPI_Log_file());

#endif

    /* New size */

    if ( MPI_SUCCESS != ( R = MMPI_Buf_unpack(
      con->recv_buffer[source],&H,1,MPI_INT) ) ) {
      MMPI_Log_abort(my_name,"Failed MMPI_Buf_unpack", R);
    }

    if ( MPI_SUCCESS != ( R = MPI_Request_free( request ) ) ) {
      MMPI_Log_abort(my_name,"Failed MPI_Request_free", R);
    }

    if ( *request != MPI_REQUEST_NULL ) {
      MMPI_Log_abort(my_name,"Failed MMPI_Request_free", MPI_ERR_REQUEST);
    }

    con->request[ source ] = *request ;

    if ( MPI_SUCCESS != ( R =
      MMPI_Buf_reset( H, con->comm, &(con->recv_buffer[ source ]) ))) {
      MMPI_Log_abort(my_name,"Failed MMPI_Buf_reset", R);
    }

    if ( MPI_SUCCESS != ( R = MMPI_Buf_recv_init(
      con->recv_buffer[ source ] ,
      status->MPI_SOURCE ,
      con->tag_con ,
      request ) ) ) {
      MMPI_Log_abort(my_name,"Failed MMPI_Buf_recv_init", R);
    }

    if ( MPI_SUCCESS != ( R = MPI_Start( request ) ) ) {
      MMPI_Log_abort(my_name,"Failed MPI_Start", R);
    }
    con->request[ source ] = *request ;

    /* Acknowledge */

    if ( MPI_SUCCESS != ( R = MPI_Rsend(NULL,0,MPI_BYTE,
      source,con->tag_ack,con->comm) ) ) {
      MMPI_Log_abort(my_name,"MPI_Rsend (ack)", R);
    }

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : from %d - Acknowledged\n",my_name,source);
  fflush(MMPI_Log_file());

#endif

    break ;

  case REQUEST_DESTROY:

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : Destroy\n",my_name);
  fflush(MMPI_Log_file());

#endif

    if ( MPI_SUCCESS != ( R = MPI_Request_free( request ) ) ) {
      MMPI_Log_abort(my_name,"Failed MPI_Request_free", R);
    }

    if ( *request != MPI_REQUEST_NULL ) {
      MMPI_Log_abort(my_name,"Failed MPI_Request_free - not null", R);
    }

    con->request[ source ] = *request ;

    if ( MPI_SUCCESS != ( R = MMPI_Buf_free( &(con->recv_buffer[source])))) {
      MMPI_Log_abort(my_name,"Failed MMPI_Buf_free", R);
    }

    /* Acknowledge */

    if ( MPI_SUCCESS != ( R = MPI_Rsend(NULL,0,MPI_BYTE,
      source,con->tag_ack,con->comm) ) ) {
      MMPI_Log_abort(my_name,"MPI_Rsend (ack)", R);
    }

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : from %d - Acknowledged\n",my_name,source);
  fflush(MMPI_Log_file());

#endif

    break ;

  case REQUEST_MESSAGE:
    {
      MMPI_Buf local_buf = MMPI_BUF_NULL ;

      MMPI_Buf_copy( con->recv_buffer[ source ] , &local_buf );

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : Message\n",my_name);
  fflush(MMPI_Log_file());

#endif

      /* OK to restart, buffer has been copied */

      if ( MPI_SUCCESS != ( R = MPI_Start( request ) ) ) {
        MMPI_Log_abort(my_name,"Failed MPI_Start", R);
      }
      con->request[ source ] = *request ;

      /* Acknowledge */

      if ( MPI_SUCCESS != ( R = MPI_Rsend(NULL,0,MPI_BYTE,
        source,con->tag_ack,con->comm) ) ) {
        MMPI_Log_abort(my_name,"MPI_Rsend (ack)", R);
      }

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : from %d - Acknowledged\n",my_name,source);
  fflush(MMPI_Log_file());

#endif

      if ( MPI_SUCCESS != ( R =
        (*(con->handler))( con->extra_state, source , local_buf ))){
        MMPI_Log_abort(my_name,"Failed consumer handler", R);
      }

      MMPI_Buf_free( &local_buf );

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : from %d - Completed\n",my_name,source);
  fflush(MMPI_Log_file());

#endif

    }

    break ;

  default:
    MMPI_Log_abort(my_name,"Undeterminable message", R);
  }

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------ */
/* The Consumer Manager creates consumers as requested */

static int con_manager(
  void        * extra_state ,
  MPI_Request * request ,
  MPI_Status  * status )
{
  const char my_name[] = "(MMPI)con_manager" ;
  int source   = status->MPI_SOURCE ;
  MMPI_Con con = (MMPI_Con) extra_state ;
  int which    = con->header[ HEADER_WHICH ];
  int maxsize  = con->header[ HEADER_SIZE ];
  int R ;

  if ( con->req_mgr != *request ) {
    MMPI_Log_abort(my_name,"Corrupted Request", MPI_ERR_REQUEST);
  }

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : from %d\n",my_name,source);
  fflush(MMPI_Log_file());

#endif

  /* Now process the request */

  switch( which ) {
  case REQUEST_CREATE:

    if ( con->request[ source ] != MPI_REQUEST_NULL ) {
      MMPI_Log_abort(my_name,"REQUEST_CREATE: Corrupted Request?",
        MPI_ERR_REQUEST);
    }

    if ( MPI_SUCCESS != ( R = MMPI_Buf_create( maxsize , con->comm ,
      &(con->recv_buffer[ source ]) ) ) ) {
      MMPI_Log_abort(my_name,"Failed MMPI_Buf_create",R);
    }

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : create %d\n",my_name,maxsize);
  fflush(MMPI_Log_file());

#endif

    if ( MPI_SUCCESS != ( R = MMPI_Buf_recv_init(
      con->recv_buffer[ source ],
      source ,
      con->tag_con ,
      &(con->request[ source ]) ) ) ) {
      MMPI_Log_abort(my_name,"Failed MMPI_Buf_recv_init",R);
    }

    if ( MPI_SUCCESS != ( R = MPI_Start( &(con->request[ source ]) ) ) ) {
      MMPI_Log_abort(my_name,"Failed MPI_Start",R);
    }

    if ( MPI_SUCCESS != ( R =
      MMPI_Post_handler( con->request[ source ], con, con_handler ) ) ) {
      MMPI_Log_abort(my_name,"Failed MMPI_Post_handler",R);
    }

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : create %d - DONE\n",my_name,maxsize);
  fflush(MMPI_Log_file());

#endif

    break ;

  default:

    MMPI_Log_abort(my_name,"Undeterminable message",MPI_ERR_UNKNOWN);
  }

  /* Acknowledge */

  if ( MPI_SUCCESS != ( R = MPI_Rsend(NULL,0,MPI_BYTE,
    source,con->tag_ack,con->comm) ) ) {
    MMPI_Log_abort(my_name,"Failed MPI_Rsend (ack)",R);
  }

  /* Receive buffer (con->header) has been processed */

  if ( MPI_SUCCESS != ( R = MPI_Start( request ) ) ) {
    MMPI_Log_abort(my_name,"Failed MPI_Start", R);
  }
  con->req_mgr = *request ;

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : from %d - Acknowledged\n",my_name,source);
  fflush(MMPI_Log_file());

#endif

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------ */

int MMPI_Con_init(
  MMPI_Con   con ,
  MMPI_Buf  * buffer ) /* IN OUT */
{
  int R ;
  int H = REQUEST_MESSAGE ;

  if ( *buffer == MMPI_BUF_NULL ) {
    if ( MPI_SUCCESS != ( R =
      MMPI_Buf_create( MINIMUM_BUFFER_SIZE , con->comm , buffer ))){
      return R ;
    }
  }
  else {
    if ( MPI_SUCCESS != ( R = MMPI_Buf_reset( 0 , con->comm , buffer ) )) {
      return R ;
    }
  }

  return MMPI_Buf_pack(&H,1,MPI_INT,buffer);
}

int MMPI_Con_send( MMPI_Buf buffer, int destination, MMPI_Con con )
{
  const char my_name[] = "MMPI_Con_send" ;
  int         sendsize ;
  int         R ;

  MMPI_Serve_handler();

  /* Verify */

  {
    MPI_Comm comm ;
    if ( MPI_SUCCESS != ( R = MMPI_Buf_comm( buffer , &comm ) ) ) {
      fprintf(MMPI_Log_file(),"%s : MMPI_Buf_comm Failed\n",my_name);
      fflush(MMPI_Log_file());
      return R ;
    }
    if ( comm != con->comm ) {
      fprintf(MMPI_Log_file(),"%s : Incompatible communicator\n",my_name);
      fflush(MMPI_Log_file());
      return MPI_ERR_COMM ;
    }
  }

  if ( MPI_SUCCESS != ( R = MMPI_Buf_position( buffer , &sendsize ) ) ) {
    fprintf(MMPI_Log_file(),"%s : MMPI_Buf_position Failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

  { /* Peek and Verify the buffer was initialized by MMPI_Con_init */
    int H ;
    int pos = 0 ;
    void * sendptr ;

    if ( MPI_SUCCESS != ( R = MMPI_Buf_pointer( buffer , &sendptr ) ) ) {
      fprintf(MMPI_Log_file(),"%s : MMPI_Buf_pointer Failed\n",my_name);
      fflush(MMPI_Log_file());
      return R ;
    }

    if (MPI_SUCCESS != (R =
      MPI_Unpack(sendptr,sendsize,&pos,&H,1,MPI_INT,con->comm))){
      fprintf(MMPI_Log_file(),"%s : MPI_Unpack Failed\n",my_name);
      fflush(MMPI_Log_file());
      return R ;
    }

    if ( H != REQUEST_MESSAGE ) {
      fprintf(MMPI_Log_file(),"%s : Bad message header\n",my_name);
      fflush(MMPI_Log_file());
      return MPI_ERR_BUFFER ;
    }
  }

  /* Check for pending message */

  if ( con->req_ack[ destination ] != MPI_REQUEST_NULL ) {
    MPI_Status status ;

    /* Wait for receiver to be ready */

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : waiting on %d\n",my_name,destination);
  fflush(MMPI_Log_file());

#endif

    if ( MPI_SUCCESS != ( R =
      MMPI_Wait( &(con->req_ack[ destination ]) , &status ) ) ) {
      fprintf(MMPI_Log_file(),"%s : MMPI_Wait failed\n",my_name);
      fflush(MMPI_Log_file());
      return R ;
    }
    if ( con->req_ack[ destination ] != MPI_REQUEST_NULL ) {
      fprintf(MMPI_Log_file(),"%s : MMPI_Wait failed ?\n",my_name);
      fflush(MMPI_Log_file());
      return MPI_ERR_UNKNOWN ;
    }

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : waiting on %d - done\n",my_name,destination);
  fflush(MMPI_Log_file());

#endif

  }

  /* Post the acknowledge receive */

  if ( MPI_SUCCESS != ( R =
    MPI_Irecv( NULL, 0, MPI_BYTE,
      destination, con->tag_ack, con->comm,
      &(con->req_ack[ destination ]) ) ) ) {
    fprintf(MMPI_Log_file(),"%s : MPI_Irecv failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

  /* Check size of destination buffer */

  if ( con->send_size[ destination ] < sendsize ) {

    MPI_Status  status ;

    /* Have to create/resize, how about a little extra while we're at it */

    sendsize += sendsize * 0.10 ; /* 10% extra */

    if ( sendsize < MINIMUM_BUFFER_SIZE ) sendsize = MINIMUM_BUFFER_SIZE ;

    if ( con->send_size[ destination ] == 0 ) {

      MPI_Request req = MPI_REQUEST_NULL ;
      int header[ HEADER_LEN ] ;

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : create con on %d\n",my_name,destination);
  fflush(MMPI_Log_file());

#endif
      header[ HEADER_WHICH ] = REQUEST_CREATE ;
      header[ HEADER_SIZE ]  = con->send_size[ destination ] = sendsize ;

      if ( MPI_SUCCESS != ( R =
        MPI_Isend( header, HEADER_LEN , MPI_INT,
          destination, con->tag_mgr, con->comm, &req ) ) ) {
        fprintf(MMPI_Log_file(),"%s : MPI_Isend failed\n",my_name);
      fflush(MMPI_Log_file());
        return R ;
      }
      /* Wait for send to clear */

      if ( MPI_SUCCESS != ( R = MMPI_Wait( &req , &status ) ) ) {
        fprintf(MMPI_Log_file(),"%s : MMPI_Wait failed\n",my_name);
      fflush(MMPI_Log_file());
        return R ;
      }
    }
    else {

      /* Resize */

      char buf[512] ;
      int  buf_pos = 0 ;
      int  H = REQUEST_RESIZE ;

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : resize con on %d\n",my_name,destination);
  fflush(MMPI_Log_file());

#endif

      if ( MPI_SUCCESS != ( R =
        MPI_Pack(&H,1,MPI_INT,buf,sizeof(buf),&buf_pos,con->comm) ) ) {
        fprintf(MMPI_Log_file(),"%s : MPI_Pack #1 failed\n",my_name);
      fflush(MMPI_Log_file());
        return R ;
      }

      con->send_size[ destination ] = sendsize ;

      if ( MPI_SUCCESS != ( R =
        MPI_Pack(&sendsize,1,MPI_INT,buf,sizeof(buf),&buf_pos,con->comm) ) ) {
        fprintf(MMPI_Log_file(),"%s : MPI_Pack #2 failed\n",my_name);
      fflush(MMPI_Log_file());
        return R ;
      }

      /* Send resize command */

      if ( MPI_SUCCESS != ( R =
        MPI_Rsend(buf,buf_pos,MPI_PACKED,destination,con->tag_con,con->comm))){
        fprintf(MMPI_Log_file(),"%s : MPI_Rsend failed\n",my_name);
      fflush(MMPI_Log_file());
        return R ;
      }
    }

    if ( MPI_SUCCESS != ( R =
      MMPI_Wait( &(con->req_ack[ destination ]) , &status ) ) ) {
      fprintf(MMPI_Log_file(),"%s : MMPI_Wait failed\n",my_name);
      fflush(MMPI_Log_file());
      return R ;
    }

    if ( con->req_ack[ destination ] != MPI_REQUEST_NULL ) {
      fprintf(MMPI_Log_file(),"%s : MMPI_Wait failed ?\n",my_name);
      fflush(MMPI_Log_file());
      return MPI_ERR_UNKNOWN ;
    }

    if ( MPI_SUCCESS != ( R =
      MPI_Irecv( NULL, 0, MPI_BYTE,
        destination, con->tag_ack, con->comm,
        &(con->req_ack[ destination ]) ) ) ) {
      fprintf(MMPI_Log_file(),"%s : MPI_Irecv failed\n",my_name);
        fflush(MMPI_Log_file());
      return R ;
    }
  }

  /* Ready to send main message */

#ifdef DEBUG

  fprintf(MMPI_Log_file(),
    "%s : sending %d BYTES to con on %d\n",my_name,sendsize,destination);
  fflush(MMPI_Log_file());

#endif

  if ( MPI_SUCCESS != ( R =
    MMPI_Buf_rsend( buffer, destination, con->tag_con ) ) ) {
    fprintf(MMPI_Log_file(),"%s : MMPI_Buf_rsend failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : Sent\n",my_name);
  fflush(MMPI_Log_file());

#endif

  /* Test for acknowlegement */

  {
    MPI_Status status ;
    int flag ;

    if ( MPI_SUCCESS != ( R =
      MMPI_Test( &(con->req_ack[ destination ]), &flag , &status ) ) ) {
      fprintf(MMPI_Log_file(),"%s : MMPI_Test failed\n",my_name);
      fflush(MMPI_Log_file());
      return R ;
    }
#ifdef DEBUG

    if ( flag ) {
      fprintf(MMPI_Log_file(),"%s : Acknowledged\n",my_name);
      fflush(MMPI_Log_file());
    }
    else {
      fprintf(MMPI_Log_file(),"%s : Acknowledgement Pending\n",my_name);
      fflush(MMPI_Log_file());
    }
#endif

  }

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------ */

int MMPI_Con_test( MMPI_Con con , int destination , int * flag )
{
  MPI_Status status ;
  MMPI_Serve_handler();
  return MMPI_Test( &(con->req_ack[ destination ]) , flag , &status );
}

int MMPI_Con_wait( MMPI_Con con , int destination )
{
  int R = MPI_SUCCESS ;
  MMPI_Serve_handler();
  while ( con->req_ack[ destination ] != MPI_REQUEST_NULL ) {
    MPI_Status status ;
    if ( MPI_SUCCESS != ( R =
      MMPI_Wait( &(con->req_ack[ destination ]) , &status ) ) ) {
      return R ;
    }
  }
  return R ;
}

/* ------------------------------------------------------------------ */

int MMPI_Con_create(
  MPI_Comm          Comm ,
  void            * extra_state ,
  MMPI_Con_handler recv_handler ,
  MMPI_Con *       con_ptr )
{
  const char my_name[] = "MMPI_Con_create" ;
  int R ;
  int i ;
  int nproc ;
  int size = 0 ;
  MMPI_Con con ;

  MMPI_Serve_handler();

  if ( MPI_SUCCESS != ( R = MPI_Comm_size( Comm , &nproc ) ) ) {
    fprintf(MMPI_Log_file(),"%s : MPI_Comm_size failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

  size += sizeof(MMPI_Con_type) ;
  size += nproc * (2*sizeof(MPI_Request) + sizeof(MMPI_Buf) + sizeof(int));

  con = *con_ptr = (MMPI_Con) malloc(size);
  con->send_size   = (int *)         ( con + 1 );
  con->recv_buffer = (MMPI_Buf *)    ( con->send_size   + nproc );
  con->request     = (MPI_Request *) ( con->recv_buffer + nproc );
  con->req_ack     = (MPI_Request *) ( con->request     + nproc );

  if ( ((char *)( con->req_ack + nproc )) - ((char *)(con)) > size ) {
    fprintf(MMPI_Log_file(),"%s : allocation foul-up\n",my_name);
      fflush(MMPI_Log_file());
    return MPI_ERR_UNKNOWN ;
  }

  con->comm     = Comm ;
  con->tag_mgr  = -1 ;
  con->tag_con = -1 ;
  con->tag_ack  = -1 ;
  con->req_mgr  = MPI_REQUEST_NULL ;
  con->handler  = recv_handler ;
  con->extra_state = extra_state ;

  for ( i = 0 ; i < nproc ; ++i ) con->send_size[i] = 0 ;
  for ( i = 0 ; i < nproc ; ++i ) con->recv_buffer[i] = MMPI_BUF_NULL ;
  for ( i = 0 ; i < nproc ; ++i ) con->request[i] = MPI_REQUEST_NULL ;
  for ( i = 0 ; i < nproc ; ++i ) con->req_ack[i] = MPI_REQUEST_NULL ;

  if ( MPI_SUCCESS != ( R = MMPI_Tag_get_global(Comm,&(con->tag_mgr)))){
    fprintf(MMPI_Log_file(),"%s : MMPI_Tag_get_global #1 failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }
  if ( MPI_SUCCESS != ( R = MMPI_Tag_get_global(Comm,&(con->tag_con)))){
    fprintf(MMPI_Log_file(),"%s : MMPI_Tag_get_global #2 failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }
  if ( MPI_SUCCESS != ( R = MMPI_Tag_get_global(Comm,&(con->tag_ack)))){
    fprintf(MMPI_Log_file(),"%s : MMPI_Tag_get_global #3 failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

  /* Create and start manager's receive */

  if ( MPI_SUCCESS != ( R = MPI_Recv_init(
    con->header, HEADER_LEN, MPI_INT,
    MPI_ANY_SOURCE, con->tag_mgr, Comm, &(con->req_mgr) ))) {
    fprintf(MMPI_Log_file(),"%s : MPI_Recv_init failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

  if ( MPI_SUCCESS != ( R = MPI_Start( &(con->req_mgr) ))){
    fprintf(MMPI_Log_file(),"%s : MPI_Start failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

  if ( MPI_SUCCESS != ( R =
    MMPI_Post_handler( con->req_mgr , con , con_manager ))) {
    fprintf(MMPI_Log_file(),"%s : MMPI_Post_handler failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------ */

int MMPI_Con_reset( MMPI_Con con )  /* Delete buffers */
{
  const char my_name[] = "MMPI_Con_reset" ;
  int nproc ;
  MPI_Status  stat ;
  int i ;
  int R ;
  int H = REQUEST_DESTROY ;
  char buf[512] ;
  int  buf_pos = 0 ;

  MMPI_Serve_handler();

  if ( MPI_SUCCESS != ( R = MPI_Comm_size( con->comm , &nproc ) ) ) {
    fprintf(MMPI_Log_file(),"%s : MPI_Comm_size failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

  if ( MPI_SUCCESS != ( R =
     MPI_Pack(&H,1,MPI_INT,buf,sizeof(buf),&buf_pos,con->comm) ) ) {
    return R ;
  }

  for ( i = 0 ; i < nproc ; ++i ) if ( con->send_size[i] ) {

    /* Check for ongoing activity */

    while ( con->req_ack[i] != MPI_REQUEST_NULL ) {

      /* Wait for receiver to be ready */

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s : waiting on %d\n",my_name,i);
  fflush(MMPI_Log_file());

#endif

      if ( MPI_SUCCESS != ( R = MMPI_Wait( &(con->req_ack[i]) , &stat ))) {
        fprintf(MMPI_Log_file(),"%s : MMPI_Wait failed\n",my_name);
        fflush(MMPI_Log_file());
        return R ;
      }
    }

    if ( MPI_SUCCESS != ( R =
      MPI_Irecv( NULL, 0, MPI_BYTE, i, con->tag_ack, con->comm,
        &(con->req_ack[i]) ) ) ) {
      fprintf(MMPI_Log_file(),"%s : MPI_Irecv failed\n",my_name);
      fflush(MMPI_Log_file());
      return R ;
    }

    if ( MPI_SUCCESS != ( R =
      MPI_Rsend( buf, buf_pos, MPI_PACKED, i, con->tag_con, con->comm) ) ) {
      return R ;
    }

    if ( MPI_SUCCESS != ( R = MMPI_Wait( &(con->req_ack[i]) , &stat ))) {
      fprintf(MMPI_Log_file(),"%s : MMPI_Wait failed\n",my_name);
      fflush(MMPI_Log_file());
      return R ;
    }
    con->send_size[i] = 0 ;
  }

  return MPI_SUCCESS ;
}

/* ------------------------------------------------------------------ */

int MMPI_Con_free( MMPI_Con * con ) /* Destroy con */
{
  const char my_name[] = "MMPI_Con_free" ;
  int R ;
  int proc , nproc , i ;
  MPI_Status  stat ;

  MMPI_Serve_handler();

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s - begin\n",my_name);
  fflush(MMPI_Log_file());

#endif

  if ( con == NULL || *con == MMPI_CON_NULL ) {
    fprintf(MMPI_Log_file(),"%s - null argument\n",my_name);
      fflush(MMPI_Log_file());
    return MPI_ERR_ARG ;
  }

  if ( MPI_SUCCESS != ( R = MPI_Comm_size( (*con)->comm , &nproc ) ) ) {
    fprintf(MMPI_Log_file(),"%s - bad communicator\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }
  if ( MPI_SUCCESS != ( R = MPI_Comm_rank( (*con)->comm , &proc ) ) ) {
    fprintf(MMPI_Log_file(),"%s - bad communicator\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

  /* Synchronize before freeing buffers */

  if ( MPI_SUCCESS != ( R = MMPI_Barrier( (*con)->comm ) ) ) {
    fprintf(MMPI_Log_file(),"%s - MMPI_Barrier Failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s - MMPI_Con_reset - BEGIN\n",my_name);
  fflush(MMPI_Log_file());

#endif

  if ( MPI_SUCCESS != ( R = MMPI_Con_reset( *con ) ) ) {
    fprintf(MMPI_Log_file(),"%s - MMPI_Con_reset Failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

  /* Wait for everyone else to reset my requests */

  for ( i = 0 ; i < nproc ; ++i ) {
    while ( MPI_REQUEST_NULL != (*con)->request[i] ) {
      MPI_Request req = (*con)->request[i] ;

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s - waiting for %d\n",my_name,i);
  fflush(MMPI_Log_file());

#endif

      if ( MPI_SUCCESS != ( R = MMPI_Wait( &req, &stat ))){
        fprintf(MMPI_Log_file(),"%s - MMPI_Wait Failed\n",my_name);
      fflush(MMPI_Log_file());
        return R ;
      }

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s - waiting for %d - done\n",my_name,i);
  fflush(MMPI_Log_file());

#endif

    }
  }

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s - MMPI_Con_reset - DONE\n",my_name);
  fflush(MMPI_Log_file());

#endif

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s - Remove handler\n",my_name);
  fflush(MMPI_Log_file());

#endif

  if ( MPI_SUCCESS != ( R = MMPI_Post_handler(
    (*con)->req_mgr , NULL , MMPI_REQUEST_HANDLER_NULL ) ) ) {
    fprintf(MMPI_Log_file(),"%s - MMPI_Post_handler Failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

  if ( MPI_SUCCESS != ( R = MPI_Rsend( NULL , 0 , MPI_INT,
    proc, (*con)->tag_mgr , (*con)->comm ) ) ) {
    fprintf(MMPI_Log_file(),"%s - MPI_Rsend Failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

  if ( MPI_SUCCESS != ( R = MPI_Wait( &((*con)->req_mgr), &stat) ) ) {
    fprintf(MMPI_Log_file(),"%s - MPI_Wait Failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

  if ( MPI_SUCCESS != ( R = MPI_Request_free( &((*con)->req_mgr) ) ) ) {
    fprintf(MMPI_Log_file(),"%s - MPI_Request_free Failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

  if ( (*con)->req_mgr != MPI_REQUEST_NULL ) {
    fprintf(MMPI_Log_file(),"%s - MPI_Request_free Failed ?\n",my_name);
      fflush(MMPI_Log_file());
    return MPI_ERR_UNKNOWN ;
  }

  /* Insure tags do not get re-used such that a message is
     sent to the killed request.
  */

  if ( MPI_SUCCESS != ( R = MMPI_Barrier( (*con)->comm ) ) ) {
    fprintf(MMPI_Log_file(),"%s - MPI_Barrier #2 Failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s - Release tags\n",my_name);
  fflush(MMPI_Log_file());

#endif

  if ( MPI_SUCCESS != ( R =
    MMPI_Tag_rel_global( (*con)->comm , &((*con)->tag_mgr) ) ) ) {
    fprintf(MMPI_Log_file(),"%s - MPI_Tag_rel_global #1 Failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }
  if ( MPI_SUCCESS != ( R =
    MMPI_Tag_rel_global( (*con)->comm , &((*con)->tag_ack) ) ) ) {
    fprintf(MMPI_Log_file(),"%s - MPI_Tag_rel_global #2 Failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }
  if ( MPI_SUCCESS != ( R =
    MMPI_Tag_rel_global( (*con)->comm , &((*con)->tag_con) ) ) ) {
    fprintf(MMPI_Log_file(),"%s - MPI_Tag_rel_global #3 Failed\n",my_name);
      fflush(MMPI_Log_file());
    return R ;
  }

  free( *con );

  *con = MMPI_CON_NULL ;

#ifdef DEBUG

  fprintf(MMPI_Log_file(),"%s - end\n",my_name);
  fflush(MMPI_Log_file());

#endif

  return MPI_SUCCESS ;
}

