

/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/






#include "PLA.h"
#include <stdio.h>
#include "MISC_DEFS.h"

/* Begin the CREATE CONFORMAL TO routines ***********************************/

/* NOTE : I don't think that Vector/Mvector conf_to are strictly correct   **/
/*        The routines do not check to see if the vector/mvector is        **/
/*        being created conformal to another (m)vector/p(m)vector.         **/
/*        Further, if you create one conformal to a projected, duplicated  **/
/*        object you do not get a duplicated object -- that is not an      **/
/*        option with (m)vectors.  This sounds minor -- but you can align  **/
/*        a duplicated object in a way that will actually wrap the indices **/
/*        so you are not assured which processor column your new (m)vector **/
/*        will start on.  This may be as intended however.                 **/

int PLA_Vector_create_conf_to(PLA_Obj   laobject,
                              PLA_Obj   *new_vector)
{
   MPI_Datatype     datatype;
   int              global_length, global_width;
   PLA_Template      template;
   int              orientation;
   int              global_align_row;
   int               global_align_col; /* Really this makes sense to me */

   int return_value = 0;

/* A vector has just a global_align_row, not a row and column                 */
/* That is -- that the number to align it to is left alone and not translated */
/* into a paired (row, column) offset.                                        */
   return_value += PLA_Obj_global_all(laobject, &global_length, &global_width,
                                      &global_align_row, &global_align_col);
   return_value += PLA_Obj_datatype(laobject, &datatype);
   return_value += PLA_Obj_template(laobject, &template);
   return_value += PLA_Obj_get_orientation(laobject, &orientation);
   if(PLA_PROJ_ONTO_ROW == orientation)
      return_value += PLA_Vector_create(datatype, global_width, template, 
                                        global_align_col, new_vector);
   else
      return_value += PLA_Vector_create(datatype, global_length, template,
                                        global_align_row, new_vector);
   return(return_value);
}

int PLA_Mvector_create_conf_to(PLA_Obj    laobject,
                               int          num_vectors,
                               PLA_Obj    *new_mvector)
{
   MPI_Datatype     datatype;
   int              global_length, global_width;
   int              orientation;
   PLA_Template     template;
   int              global_align_row, global_align_col;
   int              idummy;

   int return_value = 0;
/* I assume that a vector has just a global_align_row, not a row and column   */
/* That is -- that the number to align it to is left alone and not translated */
/* into a paired (row, column) offset.                                        */

/* Note that global_align_row is really just global_align -- the _row */
/* was used to indicate the intended slot in the inquiry routine call */

   return_value += PLA_Obj_global_all(laobject, &global_length, &global_width,
                                      &global_align_row, &global_align_col);
   return_value += PLA_Obj_datatype(laobject, &datatype);
   return_value += PLA_Obj_template(laobject, &template);
   PLA_Obj_project_onto( laobject, &orientation);
   if (orientation == PLA_PROJ_ONTO_ROW)
     return_value += PLA_Mvector_create(datatype,
                                        global_width,
                       ( num_vectors == PLA_INHERIT ? global_length : num_vectors ),
                                        template,
                                        global_align_col,
                                        new_mvector);
   else
     return_value += PLA_Mvector_create(datatype,
                                        global_length,
                       ( num_vectors == PLA_INHERIT ? global_width : num_vectors ),
                                        template,
                                        global_align_row,
                                        new_mvector);
   return(return_value);
}

int PLA_Pvector_create_conf_to(PLA_Obj   laobject,
                               int         project_onto,
                               int         owner,
                               PLA_Obj   *new_pvector)
{
   int return_value = 0;
   return_value += PLA_Pmvector_create_conf_to(laobject, project_onto,
                                               owner, 1, new_pvector);
   return(return_value);
}


int PLA_Pmvector_create_conf_to(PLA_Obj   laobject,
                                int         project_onto,
                                int         owner,
                                int         global_projected_width,
                                PLA_Obj   *new_pmvector)
{
   int              rowcol; /* type PLA_Panel_rowcol_flag */
   int return_value = 0;
   int obj_type;
   int global_length;
   int global_width;
   int global_align_row;
   int global_align_col;
   PLA_Template template;
   MPI_Datatype datatype;

   if(project_onto == PLA_PROJ_ONTO_ROW)
   {
      rowcol = PLA_PROJ_ONTO_ROW;
   }
   else if(project_onto == PLA_PROJ_ONTO_COL)
   {
      rowcol = PLA_PROJ_ONTO_COL;
   }
   else
   {
      return_value += PLA_UNKNOWN_PROJECTION;
      printf("Error in Pvector_create_conf_to -- unknown projection - exiting\n");
      return(return_value);
   }
   if(*new_pmvector != NULL) PLA_Obj_free(new_pmvector);
   
   PLA_Matrix_global_all ( laobject, 
			   &global_length, &global_width,
			   &global_align_row, &global_align_col,
			   &template);
   PLA_Obj_datatype( laobject, &datatype );
   PLA_Obj_objtype( laobject, &obj_type );

   if ( (obj_type == PLA_VECTOR) ||
	(obj_type == PLA_MVECTOR) )
     {
       if( rowcol == PLA_PROJ_ONTO_ROW)
	 return_value += PLA_Panel_create(datatype, PLA_PROJ_ONTO_ROW,
					  owner, 
					  global_projected_width,
					  global_length,
					  template,
					  0, new_pmvector);
       else
	 return_value += PLA_Panel_create(datatype, PLA_PROJ_ONTO_COL,
					  owner, 
					  global_length,
					  global_projected_width,
					  template,
					  global_align_row, new_pmvector);
     }
   else 
     {
       if( rowcol == PLA_PROJ_ONTO_ROW)
	 return_value += PLA_Panel_create(datatype, PLA_PROJ_ONTO_ROW,
					  owner, 
					  global_projected_width,
					  global_width,template,
					  global_align_col, 
					  new_pmvector);
       else
	 return_value += PLA_Panel_create(datatype, PLA_PROJ_ONTO_COL,
					  owner, 
					  global_length,
					  global_projected_width,
					  template,
					  global_align_row, new_pmvector);
     }
   return(return_value);
}

int PLA_Matrix_create_conf_to(PLA_Obj   laobject,
                              PLA_Obj   *new_matrix)
{

   int return_value = 0;
   int orientation;
   int global_width, global_length, global_align_row, global_align_col;
   PLA_Template  template = NULL;
   MPI_Datatype datatype;
   PLA_Obj_obj_type should_be_matrix;

   return_value += PLA_Obj_objtype(laobject, &should_be_matrix);
   return_value += PLA_Obj_get_orientation( laobject, &orientation);
   if(should_be_matrix != PLA_MATRIX)
   {
      return_value += DISALLOWED_OBJECT_TYPE;
      printf("Error in PLA_Matrix_create_conf_to -- a matrix can only\n");
      printf("be created CONFORMAL TO another MATRIX\n");
      return(return_value);
   }
   return_value += PLA_Obj_datatype(laobject, &datatype);
   return_value += PLA_Obj_template(laobject, &template);
   return_value += PLA_Obj_global_all(laobject, &global_length, &global_width,
                                       &global_align_row, &global_align_col);
   return_value += PLA_Matrix_create(datatype, global_length, global_width,
                                   template,global_align_row,global_align_col,
                                   new_matrix);
   return_value += PLA_Obj_set_orientation(*new_matrix, orientation);
   return(return_value);
}

int PLA_Mscalar_create_conf_to(PLA_Obj   laobject,
                               int         owner_row,
                               int         owner_col,
                               PLA_Obj   *new_mscalar)
{

   int return_value = 0;
   int global_width, global_length, idummy;
   PLA_Template  template = NULL;
   MPI_Datatype datatype;


   return_value += PLA_Obj_datatype(laobject, &datatype);
   return_value += PLA_Obj_template(laobject, &template);
   return_value += PLA_Obj_global_all(laobject, &global_length, &global_width,
                                       &idummy, &idummy);
   return_value += PLA_Obj_global_info ( laobject,
                           & global_length,
                           & global_width,
                           & idummy, 
                           ( owner_row == PLA_INHERIT ? & owner_row : & idummy ),
                           ( owner_col == PLA_INHERIT ? & owner_col : & idummy ),
                           & idummy,
                           & idummy );
   return_value += PLA_Mscalar_create(datatype, owner_row, owner_col,
				      global_length, global_width,
				      template, new_mscalar);
   return(return_value);
}


