#include "POOCLA.h"

int POOCLA_CHECKPOINTING_ACTIVE = 0;
FILE * POOCLA_CHKPT_FILE = NULL;
POOCLA_Chkpt POOCLA_CHKPT_STACK = NULL;

/*
 * checkpointing is a collective operation
 * all calls must be made collectively
 * cannot call checkpoint routines durning API active state
 */

int POOCLA_activate_checkpointing ( char * filename )
{
   int my_pe = 0;
   MPI_Comm base_comm = MPI_COMM_NULL;
   char routine_name [ 64 ] = "POOCLA_activate_checkpointing";
   extern int POOCLA_CHECKPOINTING_ACTIVE;
   extern FILE * POOCLA_CHKPT_FILE;

   if ( POOCLA_CHECKPOINTING_ACTIVE ) return ( PLA_SUCCESS );

   PLA_Base_comm_1d ( & base_comm );
   MPI_Comm_rank ( base_comm, & my_pe );

   if ( my_pe == 0 )
   {
      POOCLA_CHKPT_FILE = fopen ( filename, "ab+" );
      if ( POOCLA_CHKPT_FILE == NULL )
         PLA_Log_abort ( routine_name, "Unable to open checkpoint file",
                         PLA_ERROR_CODE, 0 );
      fseek ( POOCLA_CHKPT_FILE, 0, SEEK_SET );
   }

   POOCLA_CHECKPOINTING_ACTIVE = 1;
   return ( PLA_SUCCESS );
}

int POOCLA_deactivate_checkpointing ()
{
   char message [ 192 ];
   char routine_name [ 64 ] = "POOCLA_deactivate_checkpointing";
   extern int POOCLA_CHECKPOINTING_ACTIVE;
   extern FILE * POOCLA_CHKPT_FILE;
   POOCLA_Chkpt chkpt = NULL;

   poocla_pop_chkpt_stack ( & chkpt );

   while ( chkpt != NULL )
   {
      sprintf ( message, "Active level %d checkpoint in %s, line %d (%s) ignored\n",
                 chkpt->level, chkpt->file, chkpt->line, chkpt->comment );
      PLA_Log_message ( routine_name, message );
      poocla_pop_chkpt_stack ( & chkpt );
   }

   if ( POOCLA_CHKPT_FILE ) fclose ( POOCLA_CHKPT_FILE );
   POOCLA_CHECKPOINTING_ACTIVE = 0;
   return ( PLA_SUCCESS );
}

int poocla_set_checkpoint ( const char * const file,
                            const int line,
                            char * comment )
{
   int my_pe, cur_lvl=-1, cur_suc=PLA_FAILURE, nxt_lvl=-1, rec_len;
   fpos_t cur_pos;
   char string [ 32 ];
   POOCLA_Chkpt cur_chkpt = NULL,
                prev_chkpt = NULL;
   MPI_Comm base_comm = MPI_COMM_NULL;
   char routine_name [ 64 ] = "POOCLA_set_checkpoint";
   extern int POOCLA_CHECKPOINTING_ACTIVE;
   extern FILE * POOCLA_CHKPT_FILE;

   if ( ! POOCLA_CHECKPOINTING_ACTIVE ) return ( PLA_FAILURE );

   PLA_Base_comm_1d ( & base_comm );
   MPI_Comm_rank ( base_comm, & my_pe );

   rec_len = 3 * sizeof ( int ) + 64 * sizeof ( char );

   if ( my_pe == 0 )
   {
      fgetpos ( POOCLA_CHKPT_FILE, & cur_pos );
      fread ( & cur_lvl, sizeof ( int ), 1, POOCLA_CHKPT_FILE );
      while ( ! feof ( POOCLA_CHKPT_FILE ) )
      {
         fseek ( POOCLA_CHKPT_FILE, rec_len - sizeof ( int ), SEEK_CUR );
         if ( nxt_lvl == cur_lvl )
         {
            cur_suc = PLA_SUCCESS;
            MPI_Bcast ( & cur_suc, 1, MPI_INT, 0, base_comm );
            return ( PLA_SUCCESS );
         }
         fread ( & nxt_lvl, sizeof ( int ), 1, POOCLA_CHKPT_FILE );
      }
   }

   MPI_Bcast ( & cur_suc, 1, MPI_INT, 0, base_comm );
   if ( cur_suc == PLA_SUCCESS ) return ( PLA_SUCCESS );

/* active checkpoint */

   poocla_pop_chkpt_stack ( & prev_chkpt );

   cur_chkpt = pla_calloc ( 1, sizeof ( POOCLA_CHECKPOINT_STRUCT ) );

   cur_chkpt->level = ( prev_chkpt == NULL ? 0 : prev_chkpt->level + 1 );

   if ( cur_lvl > -1 && cur_chkpt->level != cur_lvl )
      PLA_Log_abort ( routine_name, "Incosistant checkpoints", PLA_ERROR_CODE, 0 );

   cur_suc = PLA_FAILURE;
   strncpy ( cur_chkpt->file, file, 32 );
   cur_chkpt->file [ 31 ] = '\0';
   cur_chkpt->line = line;
   strncpy ( cur_chkpt->comment, comment, 32 );
   cur_chkpt->comment [ 31 ] = '\0';

   poocla_push_chkpt_stack ( prev_chkpt );
   poocla_push_chkpt_stack ( cur_chkpt );

   if ( my_pe == 0 )
   {
      fsetpos ( POOCLA_CHKPT_FILE, & cur_pos );
      if ( cur_lvl < 0 )
      {
         fwrite ( & ( cur_chkpt->level ), sizeof ( int ), 1, POOCLA_CHKPT_FILE );
         fwrite ( & cur_suc, sizeof ( int ), 1, POOCLA_CHKPT_FILE );
         fwrite ( cur_chkpt->file, sizeof ( char ), 32, POOCLA_CHKPT_FILE );
         fwrite ( & ( cur_chkpt->line ), sizeof ( int ), 1, POOCLA_CHKPT_FILE );
         fwrite ( cur_chkpt->comment, sizeof ( char ), 32, POOCLA_CHKPT_FILE );
         fflush ( POOCLA_CHKPT_FILE );
      }
      else
         fseek ( POOCLA_CHKPT_FILE, rec_len, SEEK_CUR );
   }

   return ( PLA_FAILURE );
}

int POOCLA_end_checkpoint ( void )
{
   int my_pe, cur_suc = PLA_SUCCESS;
   fpos_t cur_pos = NULL;
   MPI_Comm base_comm = MPI_COMM_NULL;
   POOCLA_Chkpt chkpt = NULL;
   extern int POOCLA_CHECKPOINTING_ACTIVE;
   extern FILE * POOCLA_CHKPT_FILE;

   if ( ! POOCLA_CHECKPOINTING_ACTIVE ) return ( PLA_SUCCESS );

   PLA_Base_comm_1d ( & base_comm );
   MPI_Comm_rank ( base_comm, & my_pe );

   poocla_pop_chkpt_stack ( & chkpt );

   if ( my_pe == 0 )
   {
      fwrite ( & ( chkpt->level ), sizeof ( int ), 1, POOCLA_CHKPT_FILE );
      fwrite ( & cur_suc, sizeof ( int ), 1, POOCLA_CHKPT_FILE );
      fwrite ( chkpt->file, sizeof ( char ), 32, POOCLA_CHKPT_FILE );
      fwrite ( & ( chkpt->line ), sizeof ( int ), 1, POOCLA_CHKPT_FILE );
      fwrite ( chkpt->comment, sizeof ( char ), 32, POOCLA_CHKPT_FILE );
      fflush ( POOCLA_CHKPT_FILE );
   }

   pla_free ( chkpt );

   return ( PLA_SUCCESS );
}

int POOCLA_examine_chkpt_file ( char * filename )
{
   int my_pe, level, success, line;
   char file [ 32 ], comment [ 32 ];
   char routine_name [ 64 ] = "POOCLA_examine_chkpt_file";
   MPI_Comm base_comm = MPI_COMM_NULL;
   POOCLA_Chkpt chkpt = NULL;
   extern int POOCLA_CHECKPOINTING_ACTIVE;
   FILE * chkpt_file;

   if ( POOCLA_CHECKPOINTING_ACTIVE )
   {
      PLA_Log_message ( routine_name, "Checkpointing currently active" );
      return ( PLA_FAILURE );
   }

   PLA_Base_comm_1d ( & base_comm );
   MPI_Comm_rank ( base_comm, & my_pe );

   if ( my_pe == 0 )
   {
      chkpt_file = fopen ( filename, "rb" );
      if ( chkpt_file == NULL )
         PLA_Log_abort ( routine_name, "Unable to open checkpoint file",
                         PLA_ERROR_CODE, 0 );

      fread ( & level, sizeof ( int ), 1, chkpt_file );
      while ( ! feof ( chkpt_file ) )
      {
         fread ( & success, sizeof ( int ), 1, chkpt_file );
         fread ( & file, sizeof ( char ), 32, chkpt_file );
         fread ( & line, sizeof ( int ), 1, chkpt_file );
         fread ( & comment, sizeof ( char ), 32, chkpt_file );


         if ( success == PLA_SUCCESS )
            printf ( "Level %d checkpoint in %s, line %d (%s) completed\n",
                      level, file, line, comment );
         else
            printf ( "Level %d checkpoint in %s, line %d (%s) started\n",
                      level, file, line, comment );

         fflush ( stdout );

         level=-1;
         success=NULL;
         file[0]='\0';
         line=-1;
         comment[0]='\0';

         fread ( & level, sizeof ( int ), 1, chkpt_file );
      }

      fclose ( chkpt_file );
   }
   return ( PLA_SUCCESS );
}


int poocla_push_chkpt_stack ( POOCLA_Chkpt chkpt )
{
   extern POOCLA_Chkpt POOCLA_CHKPT_STACK;
   if ( POOCLA_CHKPT_STACK != NULL )
   {
      chkpt->prev_chkpt = POOCLA_CHKPT_STACK;
      POOCLA_CHKPT_STACK->next_chkpt = chkpt;
   }
   POOCLA_CHKPT_STACK = chkpt;
   return ( PLA_SUCCESS );
}

int poocla_pop_chkpt_stack ( POOCLA_Chkpt * chkpt )
{
   extern POOCLA_Chkpt POOCLA_CHKPT_STACK;
   * chkpt = POOCLA_CHKPT_STACK;
   if ( POOCLA_CHKPT_STACK != NULL )
      POOCLA_CHKPT_STACK = ( * chkpt )->prev_chkpt;
      POOCLA_CHKPT_STACK->next_chkpt == NULL;
   return ( PLA_SUCCESS );
}
