#include <mpi.h>
#include <PLA.h>
#include <MISC_DEFS.h>
#include "PRISM.h"

void PRISM_divide(PLA_Obj Scur, PLA_Obj Zcur, PLA_Obj Dcur, PLA_Obj cur_est, 
                  int cut_off, PRISM_final_list final_list)
/* 

   Scur is subproblem to work on
   Zcur is the vectors associated with Scur (with the same alignment)
   cur_est is a lower bound of the max evalue of orginal problem S
   final_list contains problems saved for end

   Assumptions upon entry:

   Original problem to be solved: 
           Given: symmetric A 
           Ultimate goal:  compute orthonormal Q s.t. Q^T A Q = D,
           a diagonal matrix.
   Status upon entry:
           Already computed: orthonormal Z = ( Z_1 Zcur Z_3 ) s.t. 
            / Z_1^T  \                        / S_1    0     0  \
            | Zcur^T | A ( Z_1 Zcur Z_2 )  =  |  0    Scur   0  |
            \ Z_3^T  /                        \  0     0    S_3 / 

   Purpose of this routine:
      if dimension of Scur is small 
          enter Scur and Zcur on final_list  to be solved later
      else
          compute ( Q_1 Q_2 ) s.t 
                / Q_1^T \ Scur ( Q_1 Q_2 ) = / Scur_1    0   \
                \ Q_2^T /                    \   0    Scur_2 /
          where the eigenvalues of Scur_1 are all greater than those of
          Scur_2.
*/
{
  int          size, length, n_0, n_half, n_1, retry, iterno, index;
  double       value;
  MPI_Datatype datatype;
  PLA_Template template = NULL;
  PLA_Obj      lambda_min = NULL, lambda_max = NULL, lambda_mean = NULL,
               Stemp = NULL, Sbeta = NULL, Ztemp = NULL, Shat_1 = NULL,
               Shat_0 = NULL, S_1 = NULL, S_0 = NULL, Zhat = NULL,
               Z_1 = NULL, Z_0 = NULL, D_1 = NULL, D_0 = NULL;
  static int recurse = 0;
  double t_start;

  /*index=PLA_Get_free_object_struct();
  if (me==ROOT)printf("index = %d \n", index); */
  
  t_start=MPI_Wtime();  
  /* # rows in Zcur */
  PLA_Obj_global_length ( Zcur, &length );
  /* # columns in Zcur which is same as size of Scur */
  PLA_Obj_global_width  ( Zcur, &size );
  PLA_Obj_datatype      ( Scur, &datatype );
  PLA_Obj_template      ( Scur, &template );

#ifdef DB_GLOBAL_
  if (me==ROOT) printf(">> PRISM_divide\n");
  PRISM_print(Scur);
  PRISM_print(Zcur);
#endif

  /* If size of Scur is small, we will compute its spectral decomposition
     as part of the end game */
  if ( size < cut_off ) { /* Step A */
    PRISM_enter_on_final_list( Scur, Zcur, Dcur, final_list );
    t_syisda[t_setup] += (MPI_Wtime()-t_start);
    return;
  }
    
  /* Compute estimates lambda_min and lambda_max so that 
     lambda_min <= lambda( S ) <= lambda_max */
  /* what if Scur not on all processors, can use PLA_ALL_* ? */
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, 
                      template, &lambda_min );
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, 
                      template, &lambda_max );
  t_syisda[t_setup] += (MPI_Wtime()-t_start);

  t_start=MPI_Wtime(); 
  /* lambda_min <= lambda(Scur) <= lambda_max */
  PRISM_range( Scur, lambda_min, lambda_max ); /* Step B */

  /* Compute the mean of the eigenvalues */
  PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, 
		     template, &lambda_mean );
  PRISM_mean( Scur, lambda_mean ); /* Step B */

  /* The max evalue of Scur is greater than lambda_mean of Scur.
     The max evalue of Scur is a lower bound of the max evalue of S.
     Thus, if lambda_mean is greater than cur_est, 
     replace with lambda_mean. */
  if (size == length)  /* Step B */
     PLA_Local_copy(lambda_mean, cur_est);  
  else if (PRISM_mscalar(cur_est) < PRISM_mscalar(lambda_mean)) 
     PLA_Local_copy(lambda_mean, cur_est); 
  
  t_syisda[t_range] += (MPI_Wtime()-t_start);

  t_start=MPI_Wtime(); 
  if ( PRISM_clustered ( lambda_min, lambda_max, cur_est ) ) { /* Step C */
    /* If lambda_min - lambda_max is small, the eigenvalues
       are in a cluster.  Take appropriate action and return */
    PRISM_deal_with_cluster( Scur, Dcur ); /* Step W */
    PLA_Obj_free( &lambda_min );
    PLA_Obj_free( &lambda_mean );
    PLA_Obj_free( &lambda_max );
    t_syisda[t_setup] += (MPI_Wtime()-t_start);
    return;
  }
  else {
    /* Stemp = Scur */

    PLA_Matrix_create_conf_to( Scur, &Stemp );
    PLA_Local_copy ( Scur, Stemp );

    /* Scale and shift Stemp so that all eigenvalues are in 
       the interval [ 0,1 ] and the mean goes to 1/2 */
    PRISM_scale_shift( Stemp, lambda_min, lambda_mean, lambda_max ); /* Step D */
    PLA_Obj_free( &lambda_min );
    PLA_Obj_free( &lambda_mean );
    PLA_Obj_free( &lambda_max );
    t_syisda[t_shift] += (MPI_Wtime()-t_start);

    t_start=MPI_Wtime();
    retry = 0;
    PLA_Matrix_create_conf_to( Scur, &Sbeta );
    PLA_Matrix_create_conf_to( Scur, &Ztemp );
    PLA_Obj_set_to_zero(Sbeta);
    PLA_Obj_set_to_zero(Ztemp);

    while (TRUE) {
      for ( iterno=0; iterno<PRISM_MAX_ITER_BETA; iterno++ ) { /* Step E,F */
	  /* apply acceleration technique before beta fn for early iterations */
	  if ( iterno<PRISM_MAX_BETA_ACCEL ) { /* Step G */
	    PRISM_accel( Stemp ); /* Step J */
	  }

	  /* apply beta fn iteration, Ztemp is a working space */ 
	  PRISM_beta( Stemp, Sbeta, Ztemp); /* Step H */

	  /* Put result back into Stemp */
	  PLA_Obj_swap( &Stemp, &Sbeta );

	  /* See if beta iteration has converged.
	     If has stop applying beta fn by breaking out of loop */
	  if ( PRISM_iter_beta_converged( Stemp, Sbeta, iterno ) ) break; /* Step I */
      }
      /* Done with Sbeta */
      PLA_Obj_free( &Sbeta );

      /* Determine # of evalues of Stemp at 0, 1/2, and 1 */
      /* should we make n_0, n_half, n_1 PLA_Obj ? */
      PRISM_count_at( Stemp, &n_0, &n_half, &n_1 ); /* Step K */

      /* If none of the evalues are at 1/2 then done iterating.
	 If all at 1/2 then cluster and don't need to iterate.
	 In either case exit loop since don't want to iterate */
      if ( n_half == 0 || n_half == size ) break; /* Step L */

      retry++;
      /* We assume that if remapping once does not finish a subproblem
	 it is an error condition.  This check stops an infinite loop
	 from unforseen circumstances */
      if ( retry > 1) {
	/* error message; ? */
	/* error return; ? */
      }

      /* Apply a function to Stemp to move the evalue at 1/2 to either
	 0 or 1 */
	PRISM_map_half( Stemp, &n_0, &n_half, &n_1 ); /* Step M */

    } /* end while */
    t_syisda[t_beta] += (MPI_Wtime()-t_start);

    t_start = MPI_Wtime();
    /* If all the evalues are at 0, 1/2, or 1 then the current subproblem
       is actually a cluster.  */
    if ( n_0 == size || n_half == size || n_1 == size ) { /* Step N */
      PRISM_deal_with_cluster( Stemp, Dcur ); /* Step U */
      PLA_Obj_free(&Stemp);
      t_syisda[t_setup] += (MPI_Wtime()-t_start);
      return;
    }
    else {
      if ( size == length ) {
	/* Only on the first divide, i.e., Stemp is S, will you get here.
	   In this case you will use Z directly since it has not had any
	   values put in it yet. */
        PLA_Obj_view_all( Zcur, &Ztemp );
      }

      /* Routine to find the orthogonal transformations Q_1 and Q_2 
	 that are put in Ztemp.  It does a reduction to tridiagonal,
	 diagonalization, and the permutes Ztemp to put all associated
	 with evalue 1 up front */
      PRISM_find_subspaces( Stemp, Ztemp ); /* Step O1, O2 */
      t_syisda[t_sbr] += (MPI_Wtime()-t_start);

      t_start = MPI_Wtime();
      /* create views for the 2 subproblems created: S_0 and S_1 */
      PLA_Obj_split_4( Stemp, n_1, n_1, &Shat_1,    PLA_DUMMY,
                                        PLA_DUMMY,  &Shat_0 );
      PLA_Matrix_create_conf_to( Shat_1, &S_1 );
      PLA_Matrix_create_conf_to( Shat_0, &S_0 );
      PLA_Obj_free( &Shat_1 );
      PLA_Obj_free( &Shat_0 );
      
      PLA_Obj_free( &Stemp );

      /* divide Scur into two new subproblems by S_1 and S_0.
	 S_1 has the range space (evalues of 1)
	 S_0 has the null space (evalues of 0) 
	 remember that S_1 and S_0 must be symmetric when done */
      PRISM_split_into_subproblems( Scur, n_1, Ztemp, S_0, S_1 ); /* Step Q,S */

      /* update the global vectors */ /* Step P */
      if ( size == length ) {
	/* Only on the first divide, i.e., Stemp is S, will you get here.
	   In this case you will use Z directly since it has not had any
	   values put in it yet. Thus, you don't need to update Z 
        PLA_Obj_swap( &Zcur, &Ztemp ); */
      }
      else {
	/* Update the global vectors Z by doing Z * Ztemp */
	/* Need temp to form product so put back in Z in end */
	PLA_Matrix_create_conf_to( Zcur, &Zhat );
	PLA_Obj_set_to_zero(Zhat);
	PLA_Gemm( PLA_NO_TRANS, PLA_NO_TRANS, PRISM_one, Zcur, Ztemp, PRISM_zero, Zhat );
	PLA_Local_copy( Zhat, Zcur );
      }
      PLA_Obj_free( &Zhat );
      PLA_Obj_free( &Ztemp );

    t_syisda[t_split] += (MPI_Wtime()-t_start);
    } /* end if on size */
    /* Create two new views of vectors that correspond to the two
       new subproblems */

    t_start=MPI_Wtime();
    PLA_Obj_vert_split_2( Zcur, n_1, &Z_1, &Z_0 );
    PLA_Obj_horz_split_2( Dcur, n_1, &D_1, &D_0 );

    recurse++;
    t_syisda[t_setup] += (MPI_Wtime()-t_start);
    
    /* Now divide the subproblem with evalues 1. */
    PRISM_divide( S_1, Z_1, D_1, cur_est, cut_off, final_list ); /* Step R */
    /* Need to decide how know when can free S_1 since may
       put on final list ? */

    PLA_Obj_free(&S_1); PLA_Obj_free(&Z_1); PLA_Obj_free(&D_1); 

    /* Now divide the subproblem with evalues 0. */
    PRISM_divide( S_0, Z_0, D_0, cur_est, cut_off, final_list ); /* Step T */
    /* Need to decide how know when can free S_0 since may
       put on final list ? */

    PLA_Obj_free(&S_0); PLA_Obj_free(&Z_0); PLA_Obj_free(&D_0); 

  } /* end if on clustered */

} /* end main */
