#include <mpi.h>
#include <PLA.h>
#include <MISC_DEFS.h>
#include "PRISM.h"

int PRISM_iter_beta_converged(PLA_Obj Sbeta, PLA_Obj Stemp, int iterno) {
/*
    Check if the beta function iterations have converged in terms 
    of the norm differences.

    Method:

    diff = ||Beta-Temp|| / ||Temp||, where ||x|| is 1-norm

    Check for convergence:
    if (diff < EPSC) 
       return TRUE;

    Check if iterations start diverging after getting close to
    convergence:
    if (interno > 0 && diff > old_diff && old_diff < EPSBKUP)
       return = TRUE

    old_diff = diff
    return FALSE
*/
    int            result;
    double         diff, n1, n2;
    void           *alpha_buffer;
    PLA_Obj        Sdiff=NULL, alpha=NULL;
    MPI_Datatype   datatype;
    PLA_Template   template;
    static double  old_diff = 1;

    /* Extract info from matrix Stemp */
    PLA_Obj_datatype( Stemp, &datatype );
    PLA_Obj_template( Stemp, &template );

    /* Temporary matrix Sdiff = Sbeta-Stemp */
    PLA_Matrix_create_conf_to( Stemp, &Sdiff ); 
    PLA_Local_copy( Sbeta, Sdiff);
    PLA_Local_axpy( PRISM_mone, Stemp, Sdiff );

    /* alpha holds the matrix norm result */
    PLA_Mscalar_create( datatype, PLA_ALL_ROWS, PLA_ALL_COLS, 1, 1, 
       template, &alpha );
    PLA_Obj_local_buffer( alpha, &alpha_buffer );

    /* Compute 1-norm of Sdiff */
    PRISM_norm1( Sdiff, alpha );
    n1 = *(double *)alpha_buffer;

    /* Compute 1-norm of Stemp */
    PRISM_norm1( Stemp, alpha );
    n2 = *(double *)alpha_buffer;

    PLA_Obj_free( &Sdiff );
    PLA_Obj_free( &alpha );

    diff = n1 / n2;
    result = FALSE;

    /* check for convergence */
    if (diff < EPSC) 
       result = TRUE;

    /* check if iterations start diverging after getting close
       to convergence */
    if (iterno && diff > old_diff && old_diff < EPSBKUP) 
       result = TRUE;

    if (result)
       old_diff = 1;
    else
       old_diff = diff;

#ifdef DB_GLOBAL_
    if (me==ROOT) {
       printf(">> PRISM_iter_beta_converged\n");
       printf("iterno=%2d diff=%12.5e\n", iterno, diff);
    }
    if (result) PRISM_print(Sbeta);
#endif

    return result;
}

